# -*- coding: utf-8 -*-
"""

    mslib.msui.socket_control
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    client socket connection handler

    This file is part of mss.

    :copyright: Copyright 2019 Shivashis Padhi
    :copyright: Copyright 2019-2021 by the mss team, see AUTHORS.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import socketio
import json
import logging

from PyQt5 import QtCore
from mslib.utils.config import MissionSupportSystemDefaultConfig as mss_default
from mslib.utils.verify_user_token import verify_user_token


class ConnectionManager(QtCore.QObject):

    signal_reload = QtCore.Signal(int, name="reload_wps")
    signal_message_receive = QtCore.Signal(str, name="message rcv")
    signal_message_reply_receive = QtCore.Signal(str, name="message reply")
    signal_message_edited = QtCore.Signal(str, name="message editted")
    signal_message_deleted = QtCore.Signal(str, name="message deleted")
    signal_new_permission = QtCore.Signal(int, int, name="new permission")
    signal_update_permission = QtCore.Signal(int, int, str, name="update permission")
    signal_revoke_permission = QtCore.Signal(int, int, name="revoke permission")
    signal_operation_permissions_updated = QtCore.Signal(int, name="operation permissions updated")
    signal_operation_list_updated = QtCore.Signal(name="operation list updated")
    signal_operation_deleted = QtCore.Signal(int, name="operation deleted")

    def __init__(self, token, user, mscolab_server_url=mss_default.mscolab_server_url):
        super(ConnectionManager, self).__init__()
        self.token = token
        self.user = user
        self.mscolab_server_url = mscolab_server_url
        if token is not None:
            logging.getLogger("engineio.client").addFilter(filter=lambda record: token not in record.getMessage())
        self.sio = socketio.Client(reconnection_attempts=5)
        self.sio.connect(self.mscolab_server_url)

        self.sio.on('file-changed', handler=self.handle_file_change)
        # on chat message recive
        self.sio.on('chat-message-client', handler=self.handle_incoming_message)
        self.sio.on('chat-message-reply-client', handler=self.handle_incoming_message_reply)
        # on message edit
        self.sio.on('edit-message-client', handler=self.handle_message_edited)
        # on message delete
        self.sio.on('delete-message-client', handler=self.handle_message_deleted)
        # on new permission
        self.sio.on('new-permission', handler=self.handle_new_permission)
        # on update of permission
        self.sio.on('update-permission', handler=self.handle_update_permission)
        # on revoking operation permission
        self.sio.on('revoke-permission', handler=self.handle_revoke_permission)
        # on updating operation permissions in admin window
        self.sio.on('operation-permissions-updated', handler=self.handle_operation_permissions_updated)
        # On Operation Delete
        self.sio.on('operation-deleted', handler=self.handle_operation_deleted)
        # On New Operation
        self.sio.on('operation-list-update', handler=self.handle_operation_list_update)

        self.sio.emit('start', {'token': token})

    def handle_update_permission(self, message):
        """
        signal update of permission affected
        """
        message = json.loads(message)
        op_id = int(message["op_id"])
        u_id = int(message["u_id"])
        access_level = message["access_level"]
        self.signal_update_permission.emit(op_id, u_id, access_level)

    def handle_new_permission(self, message):
        """
        signal updating of newly added permission
        """
        message = json.loads(message)
        op_id = int(message["op_id"])
        u_id = int(message["u_id"])
        self.signal_new_permission.emit(op_id, u_id)

    def handle_revoke_permission(self, message):
        """
        Signal update of revoked permission
        """
        message = json.loads(message)
        op_id = int(message["op_id"])
        u_id = int(message["u_id"])
        self.signal_revoke_permission.emit(op_id, u_id)

    def handle_operation_permissions_updated(self, message):
        message = json.loads(message)
        u_id = int(message["u_id"])
        self.signal_operation_permissions_updated.emit(u_id)

    def handle_incoming_message(self, message):
        # raise signal to render to view
        logging.debug(message)
        # emit signal
        self.signal_message_receive.emit(message)

    def handle_incoming_message_reply(self, message):
        self.signal_message_reply_receive.emit(message)

    def handle_message_edited(self, message):
        self.signal_message_edited.emit(message)

    def handle_message_deleted(self, message):
        self.signal_message_deleted.emit(message)

    def handle_file_change(self, message):
        message = json.loads(message)
        self.signal_reload.emit(message["op_id"])

    def handle_operation_deleted(self, message):
        op_id = int(json.loads(message)["op_id"])
        self.signal_operation_deleted.emit(op_id)

    def handle_operation_list_update(self):
        self.signal_operation_list_updated.emit()

    def handle_new_operation(self, op_id):
        logging.debug("adding user to new operation")
        self.sio.emit('add-user-to-operation', {
                      "op_id": op_id,
                      "token": self.token})

    def send_message(self, message_text, op_id, reply_id):
        if verify_user_token(self.mscolab_server_url, self.token):
            logging.debug("sending message")
            self.sio.emit('chat-message', {
                          "op_id": op_id,
                          "token": self.token,
                          "message_text": message_text,
                          "reply_id": reply_id})
        else:
            # this triggers disconnect
            self.signal_reload.emit(op_id)

    def edit_message(self, message_id, new_message_text, op_id):
        if verify_user_token(self.mscolab_server_url, self.token):
            self.sio.emit('edit-message', {
                "message_id": message_id,
                "new_message_text": new_message_text,
                "op_id": op_id,
                "token": self.token
            })
        else:
            # this triggers disconnect
            self.signal_reload.emit(op_id)

    def delete_message(self, message_id, op_id):
        if verify_user_token(self.mscolab_server_url, self.token):
            self.sio.emit('delete-message', {
                'message_id': message_id,
                'op_id': op_id,
                'token': self.token
            })
        else:
            # this triggers disconnect
            self.signal_reload.emit(op_id)

    def save_file(self, token, op_id, content, comment=None):
        # ToDo refactor API
        if verify_user_token(self.mscolab_server_url, self.token):
            logging.debug("saving file")
            self.sio.emit('file-save', {
                          "op_id": op_id,
                          "token": self.token,
                          "content": content,
                          "comment": comment})
        else:
            # this triggers disconnect
            self.signal_reload.emit(op_id)

    def disconnect(self):
        self.sio.disconnect()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'mslib/msui/ui/ui_airdata_dockwidget.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_AirdataDockwidget(object):
    def setupUi(self, AirdataDockwidget):
        AirdataDockwidget.setObjectName("AirdataDockwidget")
        AirdataDockwidget.resize(953, 106)
        self.verticalLayout = QtWidgets.QVBoxLayout(AirdataDockwidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.cbDrawAirports = QtWidgets.QCheckBox(AirdataDockwidget)
        self.cbDrawAirports.setMinimumSize(QtCore.QSize(145, 0))
        self.cbDrawAirports.setObjectName("cbDrawAirports")
        self.horizontalLayout_6.addWidget(self.cbDrawAirports)
        self.cbAirportType = CheckableComboBox(AirdataDockwidget)
        self.cbAirportType.setMinimumSize(QtCore.QSize(297, 0))
        self.cbAirportType.setLayoutDirection(QtCore.Qt.LeftToRight)
        self.cbAirportType.setObjectName("cbAirportType")
        self.horizontalLayout_6.addWidget(self.cbAirportType)
        self.btDownload = QtWidgets.QPushButton(AirdataDockwidget)
        self.btDownload.setMinimumSize(QtCore.QSize(135, 0))
        self.btDownload.setObjectName("btDownload")
        self.horizontalLayout_6.addWidget(self.btDownload)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_6.addItem(spacerItem)
        self.btApply = QtWidgets.QPushButton(AirdataDockwidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.btApply.setFont(font)
        self.btApply.setObjectName("btApply")
        self.horizontalLayout_6.addWidget(self.btApply)
        self.verticalLayout.addLayout(self.horizontalLayout_6)
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.cbDrawAirspaces = QtWidgets.QCheckBox(AirdataDockwidget)
        self.cbDrawAirspaces.setMinimumSize(QtCore.QSize(145, 0))
        self.cbDrawAirspaces.setObjectName("cbDrawAirspaces")
        self.horizontalLayout_7.addWidget(self.cbDrawAirspaces)
        self.cbAirspaces = CheckableComboBox(AirdataDockwidget)
        self.cbAirspaces.setMinimumSize(QtCore.QSize(297, 0))
        self.cbAirspaces.setObjectName("cbAirspaces")
        self.horizontalLayout_7.addWidget(self.cbAirspaces)
        self.btDownloadAsp = QtWidgets.QPushButton(AirdataDockwidget)
        self.btDownloadAsp.setMinimumSize(QtCore.QSize(135, 0))
        self.btDownloadAsp.setObjectName("btDownloadAsp")
        self.horizontalLayout_7.addWidget(self.btDownloadAsp)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_7.addItem(spacerItem1)
        self.verticalLayout.addLayout(self.horizontalLayout_7)
        self.horizontalLayout_8 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_8.setObjectName("horizontalLayout_8")
        self.cbFilterAirspaces = QtWidgets.QCheckBox(AirdataDockwidget)
        self.cbFilterAirspaces.setMinimumSize(QtCore.QSize(145, 0))
        self.cbFilterAirspaces.setObjectName("cbFilterAirspaces")
        self.horizontalLayout_8.addWidget(self.cbFilterAirspaces)
        self.label_3 = QtWidgets.QLabel(AirdataDockwidget)
        self.label_3.setObjectName("label_3")
        self.horizontalLayout_8.addWidget(self.label_3)
        self.sbFrom = QtWidgets.QDoubleSpinBox(AirdataDockwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sbFrom.sizePolicy().hasHeightForWidth())
        self.sbFrom.setSizePolicy(sizePolicy)
        self.sbFrom.setMaximumSize(QtCore.QSize(190, 16777215))
        self.sbFrom.setMaximum(100.0)
        self.sbFrom.setObjectName("sbFrom")
        self.horizontalLayout_8.addWidget(self.sbFrom)
        self.label_4 = QtWidgets.QLabel(AirdataDockwidget)
        self.label_4.setObjectName("label_4")
        self.horizontalLayout_8.addWidget(self.label_4)
        self.sbTo = QtWidgets.QDoubleSpinBox(AirdataDockwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sbTo.sizePolicy().hasHeightForWidth())
        self.sbTo.setSizePolicy(sizePolicy)
        self.sbTo.setMaximumSize(QtCore.QSize(190, 16777215))
        self.sbTo.setPrefix("")
        self.sbTo.setMaximum(100.0)
        self.sbTo.setProperty("value", 100.0)
        self.sbTo.setObjectName("sbTo")
        self.horizontalLayout_8.addWidget(self.sbTo)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_8.addItem(spacerItem2)
        self.verticalLayout.addLayout(self.horizontalLayout_8)

        self.retranslateUi(AirdataDockwidget)
        QtCore.QMetaObject.connectSlotsByName(AirdataDockwidget)

    def retranslateUi(self, AirdataDockwidget):
        _translate = QtCore.QCoreApplication.translate
        AirdataDockwidget.setWindowTitle(_translate("AirdataDockwidget", "Form"))
        self.cbDrawAirports.setText(_translate("AirdataDockwidget", "draw airports"))
        self.btDownload.setToolTip(_translate("AirdataDockwidget", "Force download or update airports"))
        self.btDownload.setText(_translate("AirdataDockwidget", "Update/Download"))
        self.btApply.setToolTip(_translate("AirdataDockwidget", "Redraw the map with the current settings"))
        self.btApply.setText(_translate("AirdataDockwidget", "Apply"))
        self.btApply.setShortcut(_translate("AirdataDockwidget", "Return"))
        self.cbDrawAirspaces.setText(_translate("AirdataDockwidget", "draw airspaces"))
        self.btDownloadAsp.setToolTip(_translate("AirdataDockwidget", "Force download or update airspaces"))
        self.btDownloadAsp.setText(_translate("AirdataDockwidget", "Update/Download"))
        self.cbFilterAirspaces.setText(_translate("AirdataDockwidget", "filter airspaces"))
        self.label_3.setText(_translate("AirdataDockwidget", "from"))
        self.sbFrom.setSuffix(_translate("AirdataDockwidget", " km"))
        self.label_4.setText(_translate("AirdataDockwidget", "to"))
        self.sbTo.setSuffix(_translate("AirdataDockwidget", " km"))
from mslib.msui.mss_qt import CheckableComboBox



from functools import partial
import re
import webbrowser

# (mss)
from PyQt5 import QtCore, QtGui, QtWidgets


TypeRole = QtCore.Qt.UserRole + 1


class DataType(object):
    """Base class for data types."""

    # (mss)
    COLOR = QtCore.Qt.black

    def matches(self, data):
        """Logic to define whether the given data matches this type."""
        raise NotImplementedError

    def next(self, model, data, parent):
        """Implement if this data type has to add child items to itself."""
        pass

    def actions(self, index):
        """Re-implement to return custom QActions."""
        return []

    def paint(self, painter, option, index):
        """Optionally re-implement for use by the delegate."""
        raise NotImplementedError

    def createEditor(self, parent, option, index):
        """Optionally re-implement for use by the delegate."""
        raise NotImplementedError

    def setModelData(self, editor, model, index):
        """Optionally re-implement for use by the delegate."""
        raise NotImplementedError

    def serialize(self, model, item, data, parent):
        """Serialize this data type."""
        value_item = parent.child(item.row(), 1)
        value = value_item.data(QtCore.Qt.DisplayRole)
        if isinstance(data, dict):
            key_item = parent.child(item.row(), 0)
            key = key_item.data(QtCore.Qt.DisplayRole)
            data[key] = value
        elif isinstance(data, list):
            data.append(value)

    def key_item(self, key, model, datatype=None, editable=True):
        """Create an item for the key column for this data type."""
        key_item = QtGui.QStandardItem(key)
        key_item.setData(datatype, TypeRole)
        key_item.setData(datatype.__class__.__name__, QtCore.Qt.ToolTipRole)
        key_item.setData(
            QtGui.QBrush(datatype.COLOR), QtCore.Qt.ForegroundRole)
        key_item.setFlags(
            QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled)
        if editable and model.editable_keys:
            key_item.setFlags(key_item.flags() | QtCore.Qt.ItemIsEditable)
        return key_item

    def value_item(self, value, model, key=None):
        """Create an item for the value column for this data type."""
        display_value = value
        item = QtGui.QStandardItem()
        item.setData(display_value, QtCore.Qt.DisplayRole)
        item.setData(value, QtCore.Qt.UserRole)
        item.setData(self, TypeRole)
        item.setData(QtGui.QBrush(self.COLOR), QtCore.Qt.ForegroundRole)
        item.setFlags(
            QtCore.Qt.ItemIsSelectable |
            QtCore.Qt.ItemIsEnabled)
        if model.editable_values:
            item.setFlags(item.flags() | QtCore.Qt.ItemIsEditable)
        return item


# -----------------------------------------------------------------------------
# Default Types
# -----------------------------------------------------------------------------


class NoneType(DataType):
    """None"""

    def matches(self, data):
        return data is None

    def value_item(self, value, model, key=None):
        item = super(NoneType, self).value_item(value, model, key)
        item.setData('None', QtCore.Qt.DisplayRole)
        return item

    def serialize(self, model, item, data, parent):
        value_item = parent.child(item.row(), 1)
        value = value_item.data(QtCore.Qt.DisplayRole)
        value = value if value != 'None' else None
        if isinstance(data, dict):
            key_item = parent.child(item.row(), 0)
            key = key_item.data(QtCore.Qt.DisplayRole)
            data[key] = value
        elif isinstance(data, list):
            data.append(value)


class StrType(DataType):
    """Strings and unicodes"""

    def matches(self, data):
        # (mss)
        return isinstance(data, str)


class IntType(DataType):
    """Integers"""

    def matches(self, data):
        return isinstance(data, int) and not isinstance(data, bool)


class FloatType(DataType):
    """Floats"""

    def matches(self, data):
        return isinstance(data, float)


class BoolType(DataType):
    """Bools are displayed as checkable items with a check box."""

    def matches(self, data):
        return isinstance(data, bool)

    def value_item(self, value, model, key=None):
        item = super(BoolType, self).value_item(value, model, key)
        item.setCheckState(QtCore.Qt.Checked if value else QtCore.Qt.Unchecked)
        item.setData('', QtCore.Qt.DisplayRole)
        if model.editable_values:
            item.setFlags(
                item.flags() | QtCore.Qt.ItemIsEditable |
                QtCore.Qt.ItemIsUserCheckable)
        return item

    def serialize(self, model, item, data, parent):
        value_item = parent.child(item.row(), 1)
        value = value_item.checkState() == QtCore.Qt.Checked
        if isinstance(data, dict):
            key_item = parent.child(item.row(), 0)
            key = key_item.data(QtCore.Qt.DisplayRole)
            data[key] = value
        elif isinstance(data, list):
            data.append(value)


class ListType(DataType):
    """Lists"""

    def matches(self, data):
        return isinstance(data, list)

    def next(self, model, data, parent):
        for i, value in enumerate(data):
            type_ = match_type(value)
            key_item = self.key_item(
                str(i), datatype=type_, editable=False, model=model)
            value_item = type_.value_item(value, model=model, key=str(i))
            parent.appendRow([key_item, value_item])
            type_.next(model, data=value, parent=key_item)

    def value_item(self, value, model, key):
        item = QtGui.QStandardItem()
        item.setFlags(QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled)
        return item

    def serialize(self, model, item, data, parent):
        key_item = parent.child(item.row(), 0)
        if key_item:
            if isinstance(data, dict):
                key = key_item.data(QtCore.Qt.DisplayRole)
                data[key] = []
                data = data[key]
            elif isinstance(data, list):
                new_data = []
                data.append(new_data)
                data = new_data
        for row in range(item.rowCount()):
            child_item = item.child(row, 0)
            type_ = child_item.data(TypeRole)
            type_.serialize(
                model=self, item=child_item, data=data, parent=item)


class DictType(DataType):
    """Dictionaries"""

    def matches(self, data):
        return isinstance(data, dict)

    def next(self, model, data, parent):
        for key, value in data.items():
            type_ = match_type(value)
            key_item = self.key_item(key, datatype=type_, model=model)
            value_item = type_.value_item(value, model, key)
            parent.appendRow([key_item, value_item])
            type_.next(model, data=value, parent=key_item)

    def value_item(self, value, model, key):
        item = QtGui.QStandardItem()
        item.setFlags(QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled)
        return item

    def serialize(self, model, item, data, parent):
        key_item = parent.child(item.row(), 0)
        if key_item:
            if isinstance(data, dict):
                key = key_item.data(QtCore.Qt.DisplayRole)
                data[key] = {}
                data = data[key]
            elif isinstance(data, list):
                new_data = {}
                data.append(new_data)
                data = new_data
        for row in range(item.rowCount()):
            child_item = item.child(row, 0)
            type_ = child_item.data(TypeRole)
            type_.serialize(model=self, item=child_item, data=data, parent=item)


# -----------------------------------------------------------------------------
# Derived Types
# -----------------------------------------------------------------------------


class RangeType(DataType):
    """A range, shown as three spinboxes next to each other.

    A range is defined as a dict with start, end and step keys.
    It supports both floats and ints.
    """

    KEYS = ['start', 'end', 'step']

    def matches(self, data):
        if isinstance(data, dict) and len(data) == 3:
            if all([True if k in self.KEYS else False for k in data.keys()]):
                return True
        return False

    def paint(self, painter, option, index):
        data = index.data(QtCore.Qt.UserRole)

        painter.save()

        painter.setPen(QtGui.QPen(index.data(QtCore.Qt.ForegroundRole).color()))
        metrics = painter.fontMetrics()
        spinbox_option = QtWidgets.QStyleOptionSpinBox()
        start_rect = QtCore.QRect(option.rect)
        start_rect.setWidth(start_rect.width() / 3.0)
        spinbox_option.rect = start_rect
        spinbox_option.frame = True
        spinbox_option.state = option.state
        spinbox_option.buttonSymbols = QtWidgets.QAbstractSpinBox.NoButtons
        for i, key in enumerate(self.KEYS):
            if i > 0:
                spinbox_option.rect.adjust(
                    spinbox_option.rect.width(), 0,
                    spinbox_option.rect.width(), 0)
            QtWidgets.QApplication.style().drawComplexControl(
                QtWidgets.QStyle.CC_SpinBox, spinbox_option, painter)
            value = str(data[key])
            value_rect = QtCore.QRectF(
                spinbox_option.rect.adjusted(6, 1, -2, -2))
            value = metrics.elidedText(
                value, QtCore.Qt.ElideRight, value_rect.width() - 20)
            painter.drawText(value_rect, value)

        painter.restore()

    def createEditor(self, parent, option, index):
        data = index.data(QtCore.Qt.UserRole)
        wid = QtWidgets.QWidget(parent)
        wid.setLayout(QtWidgets.QHBoxLayout(parent))
        wid.layout().setContentsMargins(0, 0, 0, 0)
        wid.layout().setSpacing(0)

        start = data['start']
        end = data['end']
        step = data['step']

        if isinstance(start, float):
            start_spinbox = QtWidgets.QDoubleSpinBox(wid)
        else:
            start_spinbox = QtWidgets.QSpinBox(wid)

        if isinstance(end, float):
            end_spinbox = QtWidgets.QDoubleSpinBox(wid)
        else:
            end_spinbox = QtWidgets.QSpinBox(wid)

        if isinstance(step, float):
            step_spinbox = QtWidgets.QDoubleSpinBox(wid)
        else:
            step_spinbox = QtWidgets.QSpinBox(wid)

        start_spinbox.setRange(-16777215, 16777215)
        end_spinbox.setRange(-16777215, 16777215)
        step_spinbox.setRange(-16777215, 16777215)
        start_spinbox.setValue(start)
        end_spinbox.setValue(end)
        step_spinbox.setValue(step)
        wid.layout().addWidget(start_spinbox)
        wid.layout().addWidget(end_spinbox)
        wid.layout().addWidget(step_spinbox)
        return wid

    def setModelData(self, editor, model, index):
        if isinstance(model, QtWidgets.QAbstractProxyModel):
            index = model.mapToSource(index)
            model = model.sourceModel()
        data = index.data(QtCore.Qt.UserRole)
        data['start'] = editor.layout().itemAt(0).widget().value()
        data['end'] = editor.layout().itemAt(1).widget().value()
        data['step'] = editor.layout().itemAt(2).widget().value()
        model.itemFromIndex(index).setData(data, QtCore.Qt.UserRole)

    def value_item(self, value, model, key=None):
        """Item representing a value."""
        value_item = super(RangeType, self).value_item(None, model, key)
        value_item.setData(value, QtCore.Qt.UserRole)
        return value_item

    def serialize(self, model, item, data, parent):
        value_item = parent.child(item.row(), 1)
        value = value_item.data(QtCore.Qt.UserRole)
        if isinstance(data, dict):
            key_item = parent.child(item.row(), 0)
            key = key_item.data(QtCore.Qt.DisplayRole)
            data[key] = value
        elif isinstance(data, list):
            data.append(value)


class UrlType(DataType):
    """Provide a link to urls."""

    REGEX = re.compile(r'(?:https?):\/\/|(?:file):\/\/')

    def matches(self, data):
        # (mss)
        if isinstance(data, str):
            if self.REGEX.match(data) is not None:
                return True
        return False

    def actions(self, index):
        explore = QtWidgets.QAction('Explore ...', None)
        explore.triggered.connect(
            partial(webbrowser.open, index.data(QtCore.Qt.DisplayRole)))
        return [explore]


class FilepathType(DataType):
    """Files and paths can be opened."""

    REGEX = re.compile(r'(\/.*)|([A-Z]:\\.*)')

    def matches(self, data):
        # (mss)
        if isinstance(data, str):
            if self.REGEX.match(data) is not None:
                return True
        return False

    def actions(self, index):
        explore = QtWidgets.QAction('Explore ...', None)
        path = index.data(QtCore.Qt.DisplayRole)
        explore.triggered.connect(partial(webbrowser.open, path))
        return [explore]


class ChoicesType(DataType):
    """A combobox that allows for a number of choices.

    The data has to be a dict with a value and a choices key.
    {
        "value": "A",
        "choices": ["A", "B", "C"]
    }
    """

    KEYS = ['value', 'choices']

    def matches(self, data):
        if isinstance(data, dict) and len(data) == 2:
            if all([True if k in self.KEYS else False for k in data.keys()]):
                return True
        return False

    def createEditor(self, parent, option, index):
        data = index.data(QtCore.Qt.UserRole)
        cbx = QtWidgets.QComboBox(parent)
        cbx.addItems([str(d) for d in data['choices']])
        cbx.setCurrentIndex(cbx.findText(str(data['value'])))
        return cbx

    def setModelData(self, editor, model, index):
        if isinstance(model, QtWidgets.QAbstractProxyModel):
            index = model.mapToSource(index)
            model = model.sourceModel()
        data = index.data(QtCore.Qt.UserRole)
        data['value'] = data['choices'][editor.currentIndex()]
        model.itemFromIndex(index).setData(data['value'], QtCore.Qt.DisplayRole)
        model.itemFromIndex(index).setData(data, QtCore.Qt.UserRole)

    def value_item(self, value, model, key=None):
        """Item representing a value."""
        value_item = super(ChoicesType, self).value_item(value['value'], model, key)
        value_item.setData(value, QtCore.Qt.UserRole)
        return value_item

    def serialize(self, model, item, data, parent):
        value_item = parent.child(item.row(), 1)
        value = value_item.data(QtCore.Qt.UserRole)
        if isinstance(data, dict):
            key_item = parent.child(item.row(), 0)
            key = key_item.data(QtCore.Qt.DisplayRole)
            data[key] = value
        elif isinstance(data, list):
            data.append(value)


# Add any custom DataType to this list
#
DATA_TYPES = [
    NoneType(),
    UrlType(),
    FilepathType(),
    StrType(),
    IntType(),
    FloatType(),
    BoolType(),
    ListType(),
    RangeType(),
    ChoicesType(),
    DictType()
]


def match_type(data):
    """Try to match the given data object to a DataType"""
    for type_ in DATA_TYPES:
        if type_.matches(data):
            return type_



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'mslib/msui/ui/ui_topview_window.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_TopViewWindow(object):
    def setupUi(self, TopViewWindow):
        TopViewWindow.setObjectName("TopViewWindow")
        TopViewWindow.resize(952, 782)
        self.centralwidget = QtWidgets.QWidget(TopViewWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.mpl = MplTopViewWidget(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.mpl.sizePolicy().hasHeightForWidth())
        self.mpl.setSizePolicy(sizePolicy)
        self.mpl.setMinimumSize(QtCore.QSize(100, 100))
        self.mpl.setObjectName("mpl")
        self.horizontalLayout_2.addWidget(self.mpl)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.btMapRedraw = QtWidgets.QPushButton(self.centralwidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.btMapRedraw.setFont(font)
        self.btMapRedraw.setFlat(False)
        self.btMapRedraw.setObjectName("btMapRedraw")
        self.horizontalLayout.addWidget(self.btMapRedraw)
        spacerItem = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.btSettings = QtWidgets.QPushButton(self.centralwidget)
        self.btSettings.setObjectName("btSettings")
        self.horizontalLayout.addWidget(self.btSettings)
        spacerItem1 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.btRoundtrip = QtWidgets.QPushButton(self.centralwidget)
        self.btRoundtrip.setObjectName("btRoundtrip")
        self.horizontalLayout.addWidget(self.btRoundtrip)
        spacerItem2 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem2)
        self.cbTools = QtWidgets.QComboBox(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cbTools.sizePolicy().hasHeightForWidth())
        self.cbTools.setSizePolicy(sizePolicy)
        self.cbTools.setObjectName("cbTools")
        self.cbTools.addItem("")
        self.cbTools.addItem("")
        self.horizontalLayout.addWidget(self.cbTools)
        spacerItem3 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem3)
        self.cbChangeMapSection = QtWidgets.QComboBox(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cbChangeMapSection.sizePolicy().hasHeightForWidth())
        self.cbChangeMapSection.setSizePolicy(sizePolicy)
        self.cbChangeMapSection.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)
        self.cbChangeMapSection.setObjectName("cbChangeMapSection")
        self.cbChangeMapSection.addItem("")
        self.cbChangeMapSection.addItem("")
        self.cbChangeMapSection.addItem("")
        self.cbChangeMapSection.addItem("")
        self.cbChangeMapSection.addItem("")
        self.cbChangeMapSection.addItem("")
        self.cbChangeMapSection.addItem("")
        self.horizontalLayout.addWidget(self.cbChangeMapSection)
        spacerItem4 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem4)
        self.verticalLayout.addLayout(self.horizontalLayout)
        TopViewWindow.setCentralWidget(self.centralwidget)
        self.actionCloseWindow = QtWidgets.QAction(TopViewWindow)
        self.actionCloseWindow.setObjectName("actionCloseWindow")
        TopViewWindow.addAction(self.actionCloseWindow)

        self.retranslateUi(TopViewWindow)
        self.actionCloseWindow.triggered.connect(TopViewWindow.close)
        QtCore.QMetaObject.connectSlotsByName(TopViewWindow)

    def retranslateUi(self, TopViewWindow):
        _translate = QtCore.QCoreApplication.translate
        TopViewWindow.setWindowTitle(_translate("TopViewWindow", "Top View - Mission Support System"))
        self.btMapRedraw.setToolTip(_translate("TopViewWindow", "Redraw Map (R)"))
        self.btMapRedraw.setText(_translate("TopViewWindow", "&REDRAW"))
        self.btMapRedraw.setShortcut(_translate("TopViewWindow", "R"))
        self.btSettings.setText(_translate("TopViewWindow", "options"))
        self.btRoundtrip.setText(_translate("TopViewWindow", "make roundtrip"))
        self.cbTools.setItemText(0, _translate("TopViewWindow", "(select to open tool)"))
        self.cbTools.setItemText(1, _translate("TopViewWindow", "WMS"))
        self.cbChangeMapSection.setItemText(0, _translate("TopViewWindow", "to reset map select a region"))
        self.cbChangeMapSection.setItemText(1, _translate("TopViewWindow", "Spitsbergen, large"))
        self.cbChangeMapSection.setItemText(2, _translate("TopViewWindow", "Spitsbergen, local"))
        self.cbChangeMapSection.setItemText(3, _translate("TopViewWindow", "Europe (ste)"))
        self.cbChangeMapSection.setItemText(4, _translate("TopViewWindow", "Germany (ste)"))
        self.cbChangeMapSection.setItemText(5, _translate("TopViewWindow", "Europe (cyl)"))
        self.cbChangeMapSection.setItemText(6, _translate("TopViewWindow", "Germany (cyl)"))
        self.actionCloseWindow.setText(_translate("TopViewWindow", "CloseWindow"))
        self.actionCloseWindow.setShortcut(_translate("TopViewWindow", "Ctrl+W"))
from mslib.msui.mpl_qtwidget import MplTopViewWidget



"""
    mslib.msui.multilayers
    ~~~~~~~~~~~~~~~~~~~

    This module contains classes for object oriented managing of WMS layers.
    Improves upon the old method of loading each layer on UI changes,
    the layers are all persistent and fully functional without requiring user input.

    This file is part of mss.

    :copyright: Copyright 2021 May Bär
    :copyright: Copyright 2021 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from PyQt5 import QtWidgets, QtCore, QtGui
import logging
import mslib.msui.wms_control
from mslib.msui.icons import icons
from mslib.msui.mss_qt import ui_wms_multilayers as ui
from mslib.utils.config import save_settings_qsettings, load_settings_qsettings


class Multilayers(QtWidgets.QDialog, ui.Ui_MultilayersDialog):
    """
    Contains all layers of all loaded WMS and provides helpful methods to manage them inside a popup dialog
    """

    needs_repopulate = QtCore.pyqtSignal()

    def __init__(self, dock_widget):
        super().__init__(parent=dock_widget)
        self.setupUi(self)
        self.setWindowFlags(QtCore.Qt.Window)
        if isinstance(dock_widget, mslib.msui.wms_control.HSecWMSControlWidget):
            self.setWindowTitle(self.windowTitle() + " (Top View)")
        elif isinstance(dock_widget, mslib.msui.wms_control.VSecWMSControlWidget):
            self.setWindowTitle(self.windowTitle() + " (Side View)")
        elif isinstance(dock_widget, mslib.msui.wms_control.LSecWMSControlWidget):
            self.setWindowTitle(self.windowTitle() + " (Linear View)")
        self.dock_widget = dock_widget
        self.layers = {}
        self.layers_priority = []
        self.current_layer: Layer = None
        self.threads = 0
        self.height = None
        self.scale = self.logicalDpiX() / 96
        self.filter_favourite = False
        self.carry_parameters = {"level": None, "itime": None, "vtime": None}
        self.is_linear = isinstance(dock_widget, mslib.msui.wms_control.LSecWMSControlWidget)
        self.settings = load_settings_qsettings("multilayers",
                                                {"favourites": [], "saved_styles": {}, "saved_colors": {}})
        self.synced_reference = Layer(None, None, None, is_empty=True)
        self.skip_clicked_event = False
        self.listLayers.itemChanged.connect(self.multilayer_changed)
        self.listLayers.itemClicked.connect(self.check_icon_clicked)
        self.listLayers.itemClicked.connect(self.multilayer_clicked)
        self.listLayers.itemDoubleClicked.connect(self.multilayer_doubleclicked)
        self.listLayers.setVisible(True)

        self.leMultiFilter.setVisible(True)
        self.lFilter.setVisible(True)
        self.filterFavouriteAction = self.leMultiFilter.addAction(QtGui.QIcon(icons("64x64", "star_unfilled.png")),
                                                                  QtWidgets.QLineEdit.TrailingPosition)
        self.filterRemoveAction = self.leMultiFilter.addAction(QtGui.QIcon(icons("64x64", "remove.png")),
                                                               QtWidgets.QLineEdit.TrailingPosition)
        self.filterRemoveAction.setVisible(False)
        self.filterRemoveAction.setToolTip("Click to remove the filter")
        self.filterFavouriteAction.setToolTip("Show only favourite layers")
        self.filterRemoveAction.triggered.connect(lambda x: self.remove_filter_triggered())
        self.filterFavouriteAction.triggered.connect(lambda x: self.filter_favourite_toggled())
        self.cbMultilayering.stateChanged.connect(self.toggle_multilayering)
        self.leMultiFilter.textChanged.connect(self.filter_multilayers)

        self.listLayers.setColumnWidth(2, 50)
        self.listLayers.setColumnWidth(3, 50)
        self.listLayers.setColumnWidth(1, 200)
        self.listLayers.setColumnHidden(2, True)
        self.listLayers.setColumnHidden(3, not self.is_linear)
        self.listLayers.header().setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)

        self.delete_shortcut = QtWidgets.QShortcut(QtGui.QKeySequence("Delete"), self)
        self.delete_shortcut.activated.connect(self.delete_server)
        self.delete_shortcut.setWhatsThis("Delete selected server")

    def delete_server(self, server=None):
        if not server:
            server = self.listLayers.currentItem()

        if server and not isinstance(server, Layer):
            current = self.get_current_layer()
            if current in self.layers[server.text(0)].values():
                self.current_layer = None
            for child_index in range(server.childCount()):
                widget = server.child(child_index)
                if widget in self.layers_priority:
                    self.layers_priority.remove(widget)

            index = self.listLayers.indexOfTopLevelItem(server)
            self.layers.pop(server.text(0))
            self.listLayers.takeTopLevelItem(index)
            self.update_priority_selection()
            self.needs_repopulate.emit()

    def remove_filter_triggered(self):
        self.leMultiFilter.setText("")
        if self.filter_favourite:
            self.filter_favourite_toggled()

    def filter_favourite_toggled(self):
        self.filter_favourite = not self.filter_favourite
        if self.filter_favourite:
            self.filterFavouriteAction.setIcon(QtGui.QIcon(icons("64x64", "star_filled.png")))
            self.filterFavouriteAction.setToolTip("Disable showing only favourite layers")
        else:
            self.filterFavouriteAction.setIcon(QtGui.QIcon(icons("64x64", "star_unfilled.png")))
            self.filterFavouriteAction.setToolTip("Show only favourite layers")
        self.filter_multilayers()

    def check_icon_clicked(self, item):
        """
        Checks if the mouse is pointing at an icon and handles the event accordingly
        """
        icon_width = self.height - 2

        # Clicked on layer, check favourite
        if isinstance(item, Layer):
            starts_at = 40 * self.scale
            icon_start = starts_at + 3
            if self.cbMultilayering.isChecked():
                checkbox_width = round(self.height * 0.75)
                icon_start += checkbox_width + 6
            position = self.listLayers.viewport().mapFromGlobal(QtGui.QCursor().pos())
            if icon_start <= position.x() <= icon_start + icon_width:
                self.skip_clicked_event = True
                self.threads += 1
                item.favourite_triggered()
                if self.filter_favourite:
                    self.filter_multilayers()
                self.threads -= 1

        # Clicked on server, check garbage bin
        elif isinstance(item, QtWidgets.QTreeWidgetItem):
            starts_at = 20 * self.scale
            icon_start = starts_at + 3
            position = self.listLayers.viewport().mapFromGlobal(QtGui.QCursor().pos())
            if icon_start <= position.x() <= icon_start + icon_width:
                self.threads += 1
                self.delete_server(item)
                self.threads -= 1

    def get_current_layer(self):
        """
        Return the current layer in the perspective of Multilayering or Singlelayering
        For Multilayering, it is the first priority syncable layer, or first priority layer if none are syncable
        For Singlelayering, it is the current selected layer
        """
        if self.cbMultilayering.isChecked():
            active_layers = self.get_active_layers()
            synced_layers = [layer for layer in active_layers if layer.is_synced]
            return synced_layers[0] if synced_layers else active_layers[0] if active_layers else None
        else:
            return self.current_layer

    def reload_sync(self):
        """
        Updates the self.synced_reference layer to contain the common options of all synced layers
        """
        levels, itimes, vtimes, crs = self.get_multilayer_common_options()
        self.synced_reference.levels = levels
        self.synced_reference.itimes = itimes
        self.synced_reference.vtimes = vtimes
        self.synced_reference.allowed_crs = crs

        if self.current_layer:
            if not self.synced_reference.level:
                self.synced_reference.level = self.current_layer.level
            if not self.synced_reference.itime:
                self.synced_reference.itime = self.current_layer.itime
            if not self.synced_reference.vtime:
                self.synced_reference.vtime = self.current_layer.vtime

        if self.synced_reference.level not in self.synced_reference.levels:
            self.synced_reference.level = levels[0] if levels else None

        if self.synced_reference.itime not in self.synced_reference.itimes:
            self.synced_reference.itime = itimes[-1] if itimes else None

        if self.synced_reference.vtime not in self.synced_reference.vtimes or \
                self.synced_reference.vtime < self.synced_reference.itime:
            self.synced_reference.vtime = next((vtime for vtime in vtimes if
                                                vtime >= self.synced_reference.itime), None) if vtimes else None

    def filter_multilayers(self, filter_string=None):
        """
        Hides all multilayers that don't contain the filter_string
        Shows all multilayers that do
        """
        if filter_string is None:
            filter_string = self.leMultiFilter.text()

        for wms_name in self.layers:
            header = self.layers[wms_name]["header"]
            wms_hits = 0
            for child_index in range(header.childCount()):
                widget = header.child(child_index)
                if filter_string.lower() in widget.text(0).lower() and (
                        not self.filter_favourite or widget.is_favourite):
                    widget.setHidden(False)
                    wms_hits += 1
                else:
                    widget.setHidden(True)
            if wms_hits == 0 and (len(filter_string) > 0 or self.filter_favourite):
                header.setHidden(True)
            else:
                header.setHidden(False)

        self.filterRemoveAction.setVisible(self.filter_favourite or len(filter_string) > 0)

    def get_multilayer_common_options(self, additional_layer=None):
        """
        Return the common option for levels, init_times, valid_times and CRS
        for all synchronised layers and the additional provided one
        """
        layers = self.get_active_layers(only_synced=True)
        if additional_layer:
            layers.append(additional_layer)

        elevation_values = []
        init_time_values = []
        valid_time_values = []
        crs_values = []

        for layer in layers:
            if len(layer.levels) > 0:
                elevation_values.append(layer.levels)
            init_time_values.append(layer.itimes)
            valid_time_values.append(layer.vtimes)
            crs_values.append(layer.allowed_crs)

        for values in elevation_values:
            elevation_values[0] = list(set(elevation_values[0]).intersection(values))
        for values in init_time_values:
            init_time_values[0] = list(set(init_time_values[0]).intersection(values))
        for values in valid_time_values:
            valid_time_values[0] = list(set(valid_time_values[0]).intersection(values))
        for values in crs_values:
            crs_values[0] = list(set(crs_values[0]).intersection(values))

        return sorted(elevation_values[0], key=lambda x: float(x.split()[0])) if len(elevation_values) > 0 else [], \
            sorted(init_time_values[0]) if len(init_time_values) > 0 else [], \
            sorted(valid_time_values[0]) if len(valid_time_values) > 0 else [], \
            sorted(crs_values[0]) if len(crs_values) > 0 else []

    def get_multilayer_priority(self, layer_widget):
        """
        Returns the priority of a layer, with a default of 999 if it wasn't explicitly set
        """
        priority = self.listLayers.itemWidget(layer_widget, 2)
        return int(priority.currentText()) if priority else 999

    def get_active_layers(self, only_synced=False):
        """
        Returns a list of every layer that has been checked
        """
        active_layers = []
        for wms_name in self.layers:
            header = self.layers[wms_name]["header"]
            for child_index in range(header.childCount()):
                widget = header.child(child_index)
                if widget.checkState(0) > 0 if not only_synced else widget.is_synced:
                    active_layers.append(widget)
        return sorted(active_layers, key=lambda layer: self.get_multilayer_priority(layer))

    def update_priority_selection(self):
        """
        Updates the priority numbers for the selected layers to the sorted self.layers_priority list
        """
        active_layers = self.get_active_layers()
        possible_values = [str(x) for x in range(1, len(active_layers) + 1)]
        for layer in active_layers:
            priority = self.listLayers.itemWidget(layer, 2)
            if priority is not None:
                # Update available numbers
                priority.currentIndexChanged.disconnect(self.priority_changed)
                priority.clear()
                priority.addItems(possible_values)
                # Update selected number
                priority.setCurrentIndex(self.layers_priority.index(layer))
                priority.currentIndexChanged.connect(self.priority_changed)

    def add_wms(self, wms):
        """
        Adds a wms to the multilayer list
        """
        if wms.url not in self.layers:
            header = QtWidgets.QTreeWidgetItem(self.listLayers)
            header.setText(0, wms.url)
            header.wms_name = wms.url
            self.layers[wms.url] = {}
            self.layers[wms.url]["header"] = header
            self.layers[wms.url]["wms"] = wms
            header.setExpanded(True)
            if not self.height:
                self.height = self.listLayers.visualItemRect(header).height()
            icon = QtGui.QIcon(icons("64x64", "bin.png"))
            header.setIcon(0, icon)

    def add_multilayer(self, name, wms, auto_select=False):
        """
        Adds a layer to the multilayer list, with the wms url as a parent
        """
        if name not in self.layers[wms.url]:
            layerobj = self.dock_widget.get_layer_object(wms, name.split(" | ")[-1])
            widget = Layer(self.layers[wms.url]["header"], self, layerobj, name=name)

            widget.wms_name = wms.url
            if layerobj.abstract:
                widget.setToolTip(0, layerobj.abstract)
            if self.cbMultilayering.isChecked():
                widget.setCheckState(0, QtCore.Qt.Unchecked)

            if self.is_linear:
                color = QtWidgets.QPushButton()
                color.setFixedHeight(15)
                color.setStyleSheet(f"background-color: {widget.color}")
                self.listLayers.setItemWidget(widget, 3, color)

                def color_changed(layer):
                    self.multilayer_clicked(layer)
                    new_color = QtWidgets.QColorDialog.getColor().name()
                    color.setStyleSheet(f"background-color: {new_color}")
                    layer.color_changed(new_color)
                    self.multilayer_clicked(layer)
                    self.dock_widget.auto_update()

                color.clicked.connect(lambda: color_changed(widget))

            if widget.style:
                style = QtWidgets.QComboBox()
                style.setFixedHeight(self.height)
                style.setFixedWidth(200)
                style.addItems(widget.styles)
                style.setCurrentIndex(style.findText(widget.style))

                def style_changed(layer):
                    layer.style = self.listLayers.itemWidget(layer, 1).currentText()
                    layer.style_changed()
                    self.multilayer_clicked(layer)
                    self.dock_widget.auto_update()

                style.currentIndexChanged.connect(lambda: style_changed(widget))
                self.listLayers.setItemWidget(widget, 1, style)

            size = QtCore.QSize()
            size.setHeight(self.height)
            widget.setSizeHint(0, size)

            self.layers[wms.url][name] = widget
            if widget.is_invalid:
                widget.setDisabled(True)
                return

            if not self.current_layer or auto_select:
                self.current_layer = widget
                self.listLayers.setCurrentItem(widget)

    def multilayer_clicked(self, item):
        """
        Gets called whenever the user clicks on a layer in the multilayer list
        Makes sure the dock widget updates its data depending on the users selection
        """
        if self.skip_clicked_event:
            self.skip_clicked_event = False
            return

        if not isinstance(item, Layer):
            index = self.cbWMS_URL.findText(item.text(0))
            if index != -1 and index != self.cbWMS_URL.currentIndex():
                self.cbWMS_URL.setCurrentIndex(index)
            return
        if item.is_invalid:
            return

        self.threads += 1

        if self.carry_parameters["level"] in item.get_levels():
            item.set_level(self.carry_parameters["level"])
        if self.carry_parameters["itime"] in item.get_itimes():
            item.set_itime(self.carry_parameters["itime"])
        if self.carry_parameters["vtime"] in item.get_vtimes():
            item.set_vtime(self.carry_parameters["vtime"])

        if self.current_layer != item:
            self.current_layer = item
            self.listLayers.setCurrentItem(item)
            index = self.cbWMS_URL.findText(item.get_wms().url)
            if index != -1 and index != self.cbWMS_URL.currentIndex():
                self.cbWMS_URL.setCurrentIndex(index)
            self.needs_repopulate.emit()
            if not self.cbMultilayering.isChecked():
                QtCore.QTimer.singleShot(QtWidgets.QApplication.doubleClickInterval(), self.dock_widget.auto_update)
        self.threads -= 1

    def multilayer_doubleclicked(self, item, column):
        if isinstance(item, Layer):
            self.hide()

    def multilayer_changed(self, item):
        """
        Gets called whenever the checkmark for a layer is activate or deactivated
        Creates a priority combobox or removes it depending on the situation
        """
        if self.threads > 0:
            return

        if item.checkState(0) > 0 and not self.listLayers.itemWidget(item, 2):
            priority = QtWidgets.QComboBox()
            priority.setFixedHeight(self.height)
            priority.currentIndexChanged.connect(self.priority_changed)
            self.listLayers.setItemWidget(item, 2, priority)
            self.layers_priority.append(item)
            self.update_priority_selection()
            if (item.itimes or item.vtimes or item.levels) and self.is_sync_possible(item):
                item.is_synced = True
                self.reload_sync()
            elif not (item.itimes or item.vtimes or item.levels):
                item.is_active_unsynced = True
            self.update_checkboxes()
            self.needs_repopulate.emit()
            self.dock_widget.auto_update()
        elif item.checkState(0) == 0 and self.listLayers.itemWidget(item, 2):
            if item in self.layers_priority:
                self.listLayers.removeItemWidget(item, 2)
                self.layers_priority.remove(item)
                self.update_priority_selection()
            item.is_synced = False
            item.is_active_unsynced = False
            self.reload_sync()
            self.update_checkboxes()
            self.needs_repopulate.emit()
            self.dock_widget.auto_update()

    def priority_changed(self, new_index):
        """
        Get called whenever the user changes a priority for a layer
        Finds out the previous index and switches the layer position in self.layers_priority
        """
        active_layers = self.get_active_layers()
        old_index = [i for i in range(1, len(active_layers) + 1)]
        for layer in active_layers:
            value = self.get_multilayer_priority(layer)
            if value in old_index:
                old_index.remove(value)
        old_index = old_index[0] - 1

        to_move = self.layers_priority.pop(old_index)
        self.layers_priority.insert(new_index, to_move)
        self.update_priority_selection()
        self.multilayer_clicked(self.layers_priority[new_index])
        self.needs_repopulate.emit()
        self.dock_widget.auto_update()

    def update_checkboxes(self):
        """
        Activates or deactivates the checkboxes for every layer depending on whether they
        can be synchronised or not
        """
        self.threads += 1
        for wms_name in self.layers:
            header = self.layers[wms_name]["header"]
            for child_index in range(header.childCount()):
                layer = header.child(child_index)
                is_active = self.is_sync_possible(layer) or not (layer.itimes or layer.vtimes or layer.levels)
                layer.setDisabled(not is_active or layer.is_invalid)
        self.threads -= 1

    def is_sync_possible(self, layer):
        """
        Returns whether the passed layer can be synchronised with all other synchronised layers
        """
        if len(self.get_active_layers()) == 0:
            return True

        levels, itimes, vtimes, crs = self.get_multilayer_common_options(layer)
        levels_before, itimes_before, vtimes_before, crs_before = self.get_multilayer_common_options()

        return (len(levels) > 0 or (len(levels_before) == 0 and len(layer.levels) == 0)) and \
               (len(itimes) > 0 or (len(itimes_before) == 0 and len(layer.itimes) == 0)) and \
               (len(vtimes) > 0 or (len(vtimes_before) == 0 and len(layer.vtimes) == 0))

    def toggle_multilayering(self):
        """
        Toggle between checkable layers (multilayering) and single layer mode
        """
        self.threads += 1
        for wms_name in self.layers:
            header = self.layers[wms_name]["header"]
            for child_index in range(header.childCount()):
                layer = header.child(child_index)
                if self.cbMultilayering.isChecked():
                    layer.setCheckState(0, 2 if layer.is_synced or layer.is_active_unsynced else 0)
                else:
                    layer.setData(0, QtCore.Qt.CheckStateRole, QtCore.QVariant())
                    layer.setDisabled(layer.is_invalid)

        if self.cbMultilayering.isChecked():
            self.update_checkboxes()
            self.listLayers.setColumnHidden(2, False)
        else:
            self.listLayers.setColumnHidden(2, True)

        self.needs_repopulate.emit()
        self.threads -= 1


class Layer(QtWidgets.QTreeWidgetItem):
    def __init__(self, header, parent, layerobj, name=None, is_empty=False):
        super().__init__(header)
        self.parent = parent
        self.header = header
        self.layerobj = layerobj
        self.dimensions = {}
        self.extents = {}
        self.setText(0, name if name else "")

        self.levels = []
        self.level = None
        self.itimes = []
        self.itime = None
        self.itime_name = None
        self.allowed_init_times = []
        self.vtimes = []
        self.vtime = None
        self.vtime_name = None
        self.allowed_valid_times = []
        self.styles = []
        self.style = None
        self.is_synced = False
        self.is_active_unsynced = False
        self.is_favourite = False
        self.is_invalid = False

        if not is_empty:
            self._parse_layerobj()
            self._parse_levels()
            self._parse_itimes()
            self._parse_vtimes()
            self._parse_styles()
            self.is_favourite = str(self) in self.parent.settings["favourites"]
            self.show_favourite()
            if str(self) in self.parent.settings["saved_colors"]:
                self.color = self.parent.settings["saved_colors"][str(self)]
            else:
                self.color = "#00aaff"

    def _parse_layerobj(self):
        """
        Parses the dimensions and extents out of the self.layerobj
        """
        self.allowed_crs = []
        lobj = self.layerobj
        while lobj is not None:
            self.dimensions.update(lobj.dimensions)
            for key in lobj.extents:
                if key not in self.extents:
                    self.extents[key] = lobj.extents[key]
            if len(self.allowed_crs) == 0:
                self.allowed_crs = getattr(lobj, "crsOptions", None)
            lobj = lobj.parent

    def _parse_levels(self):
        """
        Extracts and saves the possible levels for the layer
        """
        if "elevation" in self.extents:
            units = self.dimensions["elevation"]["units"]
            values = self.extents["elevation"]["values"]
            self.levels = [f"{e.strip()} ({units})" for e in values]
            self.level = self.levels[0]

    def _parse_itimes(self):
        """
        Extracts and saves all init_time values for the layer
        """
        init_time_names = [x for x in ["init_time", "reference_time", "run"] if x in self.extents]

        # Both time dimension and time extent tags were found. Try to determine the
        # format of the date/time strings.
        if len(init_time_names) > 0:
            self.itime_name = init_time_names[0]
            values = self.extents[self.itime_name]["values"]
            self.allowed_init_times = sorted(self.parent.dock_widget.parse_time_extent(values))
            self.itimes = [_time.isoformat() + "Z" for _time in self.allowed_init_times]
            if len(self.allowed_init_times) == 0:
                logging.error(f"Cannot determine init time format of {self.header.text(0)} for {self.text(0)}")
                self.is_invalid = True
            else:
                self.itime = self.itimes[-1]

    def _parse_vtimes(self):
        """
        Extracts and saves all valid_time values for the layer
        """
        valid_time_names = [x for x in ["time", "forecast"] if x in self.extents]

        # Both time dimension and time extent tags were found. Try to determine the
        # format of the date/time strings.
        if len(valid_time_names) > 0:
            self.vtime_name = valid_time_names[0]
            values = self.extents[self.vtime_name]["values"]
            self.allowed_valid_times = sorted(self.parent.dock_widget.parse_time_extent(values))
            self.vtimes = [_time.isoformat() + "Z" for _time in self.allowed_valid_times]
            if len(self.allowed_valid_times) == 0:
                logging.error(f"Cannot determine valid time format of {self.header.text(0)} for {self.text(0)}")
                self.is_invalid = True
            else:
                if self.itime:
                    self.vtime = next((vtime for vtime in self.vtimes if vtime >= self.itime), self.vtimes[0])
                else:
                    self.vtime = self.vtimes[0]

    def _parse_styles(self):
        """
        Extracts and saves all styles for the layer.
        Sets the layers style to the first one, or the saved one if possible.
        """
        self.styles = [f"{style} | {self.layerobj.styles[style]['title']}" for style in self.layerobj.styles]
        if self.parent.is_linear:
            self.styles.extend(["linear | linear scaled y-axis", "log | log scaled y-axis"])
        if len(self.styles) > 0:
            self.style = self.styles[0]
            if str(self) in self.parent.settings["saved_styles"] and \
               self.parent.settings["saved_styles"][str(self)] in self.styles:
                self.style = self.parent.settings["saved_styles"][str(self)]

    def get_level(self):
        if not self.parent.cbMultilayering.isChecked() or not self.is_synced:
            return self.level
        else:
            return self.parent.synced_reference.level

    def get_levels(self):
        if not self.parent.cbMultilayering.isChecked() or not self.is_synced:
            return self.levels
        else:
            return self.parent.synced_reference.levels

    def get_itimes(self):
        if not self.parent.cbMultilayering.isChecked() or not self.is_synced:
            return self.itimes
        else:
            return self.parent.synced_reference.itimes

    def get_itime(self):
        if not self.parent.cbMultilayering.isChecked() or not self.is_synced:
            return self.itime
        else:
            return self.parent.synced_reference.itime

    def get_vtimes(self):
        if not self.parent.cbMultilayering.isChecked() or not self.is_synced:
            return self.vtimes
        else:
            return self.parent.synced_reference.vtimes

    def get_vtime(self):
        if not self.parent.cbMultilayering.isChecked() or not self.is_synced:
            return self.vtime
        else:
            return self.parent.synced_reference.vtime

    def set_level(self, level):
        if (not self.parent.cbMultilayering.isChecked() or not self.is_synced) and level in self.levels:
            self.level = level
        elif self.is_synced and level in self.parent.synced_reference.levels:
            self.parent.synced_reference.level = level

    def set_itime(self, itime):
        if (not self.parent.cbMultilayering.isChecked() or not self.is_synced) and itime in self.itimes:
            self.itime = itime
        elif self.is_synced and itime in self.parent.synced_reference.itimes:
            self.parent.synced_reference.itime = itime

        if self.get_vtime():
            if self.get_vtime() < itime:
                valid_vtime = next((vtime for vtime in self.get_vtimes() if vtime >= itime), None)
                if valid_vtime:
                    self.set_vtime(valid_vtime)
                    self.parent.carry_parameters["vtime"] = self.get_vtime()
            self.parent.needs_repopulate.emit()

    def set_vtime(self, vtime):
        if (not self.parent.cbMultilayering.isChecked() or not self.is_synced) and vtime in self.vtimes:
            self.vtime = vtime
        elif self.is_synced and vtime in self.parent.synced_reference.vtimes:
            self.parent.synced_reference.vtime = vtime

        if self.get_itime() and self.get_itime() > vtime:
            valid_itimes = [itime for itime in self.get_itimes() if itime <= vtime]
            if valid_itimes:
                self.set_itime(valid_itimes[-1])
                self.parent.needs_repopulate.emit()

    def get_layer(self):
        """
        Returns the layer name used internally by the WMS
        """
        return self.text(0).split(" | ")[-1].split(" (synced)")[0]

    def get_style(self):
        """
        Returns the style name used internally by the WMS
        """
        if self.style:
            return self.style.split(" |")[0]
        return ""

    def get_level_name(self):
        """
        Returns the level used internally by the WMS
        """
        if self.level:
            return self.get_level().split(" (")[0]

    def get_legend_url(self):
        if not self.parent.is_linear:
            style = self.get_style()
            urlstr = None
            if style and "legend" in self.layerobj.styles[style]:
                urlstr = self.layerobj.styles[style]["legend"]
            return urlstr

    def get_allowed_crs(self):
        if self.is_synced:
            return self.parent.synced_reference.allowed_crs
        else:
            return self.allowed_crs

    def draw(self):
        """
        Triggers the layer to be drawn by the WMSControlWidget
        """
        if isinstance(self.parent.dock_widget, mslib.msui.wms_control.HSecWMSControlWidget):
            self.parent.dock_widget.get_map([self])
        elif isinstance(self.parent.dock_widget, mslib.msui.wms_control.VSecWMSControlWidget):
            self.parent.dock_widget.get_vsec([self])
        else:
            self.parent.dock_widget.get_lsec([self])

    def get_wms(self):
        return self.parent.layers[self.header.text(0)]["wms"]

    def show_favourite(self):
        """
        Shows a filled star icon if this layer is a favourite layer or an unfilled one if not
        """
        if self.is_favourite:
            icon = QtGui.QIcon(icons("64x64", "star_filled.png"))
        else:
            icon = QtGui.QIcon(icons("64x64", "star_unfilled.png"))
        self.setIcon(0, icon)

    def style_changed(self):
        """
        Persistently saves the currently selected style of the layer, if it is not the first one
        """
        if self.style != self.styles[0]:
            self.parent.settings["saved_styles"][str(self)] = self.style
        else:
            self.parent.settings["saved_styles"].pop(str(self))
        save_settings_qsettings("multilayers", self.parent.settings)

    def color_changed(self, color):
        """
        Persistently saves the currently selected color of the layer, if it isn't black
        """
        self.color = color
        if self.color != 0:
            self.parent.settings["saved_colors"][str(self)] = self.color
        else:
            self.parent.settings["saved_colors"].pop(str(self))
        save_settings_qsettings("multilayers", self.parent.settings)

    def favourite_triggered(self):
        """
        Toggles whether a layer is or is not a favourite
        """
        self.is_favourite = not self.is_favourite
        self.show_favourite()
        if not self.is_favourite and str(self) in self.parent.settings["favourites"]:
            self.parent.settings["favourites"].remove(str(self))
        elif self.is_favourite and str(self) not in self.parent.settings["favourites"]:
            self.parent.settings["favourites"].append(str(self))
        save_settings_qsettings("multilayers", self.parent.settings)

    def __str__(self):
        return f"{self.header.text(0) if self.header else ''}: {self.text(0)}"



# -*- coding: utf-8 -*-
"""

    mslib.plugins.io.text
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    plugin for text format flight track export

    This file is part of mss.

    :copyright: Copyright 2008-2014 Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
    :copyright: Copyright 2011-2014 Marc Rautenhaus (mr)
    :copyright: Copyright 2016-2021 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import division

from builtins import str

import logging
import codecs
import os
from fs import open_fs

import mslib.msui.flighttrack as ft
from mslib.utils.units import units
from mslib.utils import thermolib


def save_to_txt(filename, name, waypoints):
    if not filename:
        raise ValueError("filename to save flight track cannot be None")
    max_loc_len, max_com_len = len("Location"), len("Comments")
    for wp in waypoints:
        if len(str(wp.location)) > max_loc_len:
            max_loc_len = len(str(wp.location))
        if len(str(wp.comments)) > max_com_len:
            max_com_len = len(str(wp.comments))
    with codecs.open(filename, "w", encoding="utf-8") as out_file:
        out_file.write(u"# Do not modify if you plan to import this file again!\n")
        out_file.write(f"Track name: {name:}\n")
        line = u"{0:5d}  {1:{2}}  {3:10.3f}  {4:11.3f}  {5:11.3f}  {6:14.3f}  {7:14.1f}  {8:15.1f}  {9:{10}}\n"
        header = f"Index  {'Location':{max_loc_len}}  Lat (+-90)  Lon (+-180)  Flightlevel  Pressure (hPa)  " \
                 f"Leg dist. (km)  Cum. dist. (km)  {'Comments':{max_com_len}}\n"
        out_file.write(header)
        for i, wp in enumerate(waypoints):
            # ToDo check str(str( .. ) and may be use csv write
            loc = str(wp.location)
            lat = wp.lat
            lon = wp.lon
            lvl = wp.flightlevel
            pre = wp.pressure / 100.
            leg = wp.distance_to_prev
            cum = wp.distance_total
            com = str(wp.comments)
            out_file.write(line.format(i, loc, max_loc_len, lat, lon, lvl, pre, leg, cum, com, max_com_len))


def load_from_txt(filename):
    name = os.path.basename(filename.replace(".txt", "").strip())
    waypoints = []
    _dirname, _name = os.path.split(filename)
    _fs = open_fs(_dirname)
    with _fs.open(_name, "r") as in_file:
        pos = []
        for line in in_file:
            if line.startswith("#"):
                continue
            if line.startswith("Index"):
                pos.append(0)  # 0
                pos.append(line.find("Location"))  # 1
                pos.append(line.find("Lat (+-90)"))  # 2
                pos.append(line.find("Lon (+-180)"))  # 3
                pos.append(line.find("Flightlevel"))  # 4
                pos.append(line.find("Pressure (hPa)"))  # 5
                pos.append(line.find("Leg dist. (km)"))  # 6
                pos.append(line.find("Cum. dist. (km)"))  # 7
                pos.append(line.find("Comments"))  # 8
                continue
            if line.startswith("Track name: "):
                continue

            if len(pos) == 0:
                raise SyntaxError("TXT Import could not parse column headings.")
            if len(line) < max(pos):
                raise SyntaxError(f"TXT Import could not parse line: '{line}'")

            wp = ft.Waypoint()
            attr_names = ["location", "lat", "lon", "flightlevel",
                          "pressure", "distance_to_prev", "distance_total",
                          "comments"]
            setattr(wp, attr_names[0], line[pos[1]:pos[2]].strip())
            setattr(wp, attr_names[7], line[pos[8]:].strip())
            for i in range(2, len(pos) - 1):
                if pos[i] >= 0:
                    if i == 5:
                        setattr(wp, attr_names[i - 1],
                                float(line[pos[i]:pos[i + 1]].strip()) * 100.)
                    else:
                        setattr(wp, attr_names[i - 1],
                                float(line[pos[i]:pos[i + 1]].strip()))
                else:
                    if i == 5:
                        logging.debug('calculate pressure from FL ' + str(
                            thermolib.flightlevel2pressure(float(wp.flightlevel) * units.hft).magnitude))
                        setattr(wp, attr_names[i - 1],
                                thermolib.flightlevel2pressure(float(wp.flightlevel) * units.hft).magnitude)
            waypoints.append(wp)
    return name, waypoints



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui_wms_password_dialog.ui'
#
# Created by: PyQt5 UI code generator 5.6
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_WMSAuthenticationDialog(object):
    def setupUi(self, WMSAuthenticationDialog):
        WMSAuthenticationDialog.setObjectName("WMSAuthenticationDialog")
        WMSAuthenticationDialog.resize(493, 151)
        self.verticalLayout = QtWidgets.QVBoxLayout(WMSAuthenticationDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.lblMessage = QtWidgets.QLabel(WMSAuthenticationDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.lblMessage.setFont(font)
        self.lblMessage.setObjectName("lblMessage")
        self.verticalLayout.addWidget(self.lblMessage)
        self.label = QtWidgets.QLabel(WMSAuthenticationDialog)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label_2 = QtWidgets.QLabel(WMSAuthenticationDialog)
        self.label_2.setMinimumSize(QtCore.QSize(80, 0))
        self.label_2.setMaximumSize(QtCore.QSize(80, 16777215))
        self.label_2.setObjectName("label_2")
        self.horizontalLayout.addWidget(self.label_2)
        self.leUsername = QtWidgets.QLineEdit(WMSAuthenticationDialog)
        self.leUsername.setObjectName("leUsername")
        self.horizontalLayout.addWidget(self.leUsername)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label_3 = QtWidgets.QLabel(WMSAuthenticationDialog)
        self.label_3.setMinimumSize(QtCore.QSize(80, 0))
        self.label_3.setMaximumSize(QtCore.QSize(80, 16777215))
        self.label_3.setObjectName("label_3")
        self.horizontalLayout_2.addWidget(self.label_3)
        self.lePassword = QtWidgets.QLineEdit(WMSAuthenticationDialog)
        self.lePassword.setInputMask("")
        self.lePassword.setText("")
        self.lePassword.setEchoMode(QtWidgets.QLineEdit.Password)
        self.lePassword.setObjectName("lePassword")
        self.horizontalLayout_2.addWidget(self.lePassword)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.buttonBox = QtWidgets.QDialogButtonBox(WMSAuthenticationDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(WMSAuthenticationDialog)
        self.buttonBox.accepted.connect(WMSAuthenticationDialog.accept)
        self.buttonBox.rejected.connect(WMSAuthenticationDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(WMSAuthenticationDialog)

    def retranslateUi(self, WMSAuthenticationDialog):
        _translate = QtCore.QCoreApplication.translate
        WMSAuthenticationDialog.setWindowTitle(_translate("WMSAuthenticationDialog", "HTTP Authentication"))
        self.lblMessage.setText(_translate("WMSAuthenticationDialog", "Web Map Service"))
        self.label.setText(_translate("WMSAuthenticationDialog", "The server you are trying to connect requires a username and a password:"))
        self.label_2.setText(_translate("WMSAuthenticationDialog", "User name:"))
        self.label_3.setText(_translate("WMSAuthenticationDialog", "Password:"))




# -*- coding: utf-8 -*-
"""

    mslib.msui.wms_control
    ~~~~~~~~~~~~~~~~~~~~~~

    Control widget to access Web Map Services.

    This file is part of mss.

    :copyright: Copyright 2008-2014 Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
    :copyright: Copyright 2011-2014 Marc Rautenhaus (mr)
    :copyright: Copyright 2016-2021 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import time
from datetime import datetime

import io
import hashlib
import logging
import mpl_toolkits.basemap as basemap
import os
import requests
import traceback
import urllib.parse
import defusedxml.ElementTree as etree
from mslib.utils.config import config_loader, load_settings_qsettings, save_settings_qsettings
from PyQt5 import QtCore, QtGui, QtWidgets

import owslib.util
from owslib.crs import axisorder_yx
from PIL import Image, ImageOps

from mslib.msui import constants, wms_capabilities
from mslib.msui.mss_qt import ui_wms_dockwidget as ui
from mslib.msui.mss_qt import ui_wms_password_dialog as ui_pw
from mslib.msui.mss_qt import Worker
from mslib.msui.multilayers import Multilayers, Layer
import mslib.utils.ogcwms as ogcwms
from mslib.utils.time import parse_iso_datetime, parse_iso_duration


WMS_SERVICE_CACHE = {}
WMS_URL_LIST = QtGui.QStandardItemModel()


def add_wms_urls(combo_box, url_list):
    combo_box_urls = [combo_box.itemText(_i) for _i in range(combo_box.count())]
    for url in (_url for _url in url_list if _url not in combo_box_urls):
        combo_box.addItem(url)


class MSSWebMapService(ogcwms.WebMapService):
    """Overloads the getmap() method of owslib.wms.WebMapService:

        added parameters are
         init_time
         level
         path_str

        the function times out after wms_timeout seconds.
    """

    def getmap(self, layers=None, styles=None, srs=None, bbox=None,
               format=None, size=None, time=None, init_time=None,
               path_str=None, level=None, transparent=False, bgcolor='#FFFFFF',
               time_name="time", init_time_name="init_time",
               exceptions='XML', method='Get',
               return_only_url=False):
        """Request and return an image from the WMS as a file-like object.

        Parameters
        ----------
        layers : list
            List of content layer names.
        styles : list
            Optional list of named styles, must be the same length as the
            layers list.
        srs : string
            A spatial reference system identifier.
        bbox : tuple
            (left, bottom, right, top) in srs units.
        format : string
            Output image format such as 'image/jpeg'.
        size : tuple
            (width, height) in pixels.
        transparent : bool
            Optional. Transparent background if True.
        bgcolor : string
            Optional. Image background color.
        method : string
            Optional. HTTP DCP method name: Get or Post.

        Example
        -------
            >>> img = wms.getmap(layers=['global_mosaic'],
            ...                  styles=['visual'],
            ...                  srs='EPSG:4326',
            ...                  bbox=(-112,36,-106,41),
            ...                  format='image/jpeg',
            ...                  size=(300,250),
            ...                  transparent=True,
            ...                  )
            >>> out = open('example.jpg', 'wb')
            >>> out.write(img.read())
            >>> out.close()

        """
        base_url = self.get_redirect_url(method)
        request = {'version': self.version, 'request': 'GetMap'}

        if self.version != "1.3.0":
            exceptions = "application/vnd.ogc.se_xml"

        # check layers and styles
        assert len(layers) > 0
        request['layers'] = ','.join(layers)
        if styles:
            assert len(styles) == len(layers)
            request['styles'] = ','.join(styles)
        else:
            request['styles'] = ''

        # size
        request['width'] = str(size[0])
        request['height'] = str(size[1])

        request['srs' if self.version != "1.3.0" else "crs"] = str(srs)
        request['format'] = str(format)
        request['transparent'] = str(transparent).upper()
        request['bgcolor'] = '0x' + bgcolor[1:7]
        request['exceptions'] = str(exceptions)

        # ++(mss)
        if bbox is not None:
            request['bbox'] = ','.join([str(x) for x in bbox])
        if path_str is not None:
            request['path'] = path_str

        # According to the WMS 1.1.1 standard, dimension names must be
        # prefixed with a "dim_", except for the in the standard predefined
        # dimensions "time" and "elevation".
        if time is not None:
            time_name = time_name if time_name == "time" else "dim_" + time_name
        if init_time is not None:
            init_time_name = "dim_" + init_time_name

        # If the parameters time and init_time are given as datetime objects,
        # create formatted strings with the given formatter. If they are
        # given as strings, use these strings directly as WMS arguments.
        if isinstance(time, datetime):
            request[time_name] = time.isoformat() + "Z"
        elif isinstance(time, str):
            request[time_name] = time

        if isinstance(init_time, datetime):
            request[init_time_name] = init_time.isoformat() + "Z"
        elif isinstance(init_time, str):
            request[init_time_name] = init_time

        if level is not None:
            request['elevation'] = str(level)

        # normalise base_url so it contains no request and no parameters
        (scheme, netloc, path, params, query, fragment) = urllib.parse.urlparse(base_url)
        base_url = urllib.parse.urlunparse((scheme, netloc, path, params, "", fragment))
        request.update(dict(urllib.parse.parse_qsl(query)))
        # --(mss)

        data = urllib.parse.urlencode(request)

        # ++(mss)
        base_url = base_url.replace("ogctest.iblsoft", "ogcie.iblsoft")  # IBL Bugfix!
        base_url = base_url.replace("ogcie/obs", "ogcie.iblsoft.com/obs")  # IBL Bugfix!
        base_url = base_url.replace(", staging1", "")  # geo.beopen.eu bugfix

        complete_url = urllib.parse.urlunparse((str(""), str(""), str(base_url), str(""), data, str("")))
        if return_only_url:
            return complete_url
        logging.debug("Retrieving image from '%s'", complete_url)
        # --(mss)

        # (mss) owslib.util.openURL checks for ServiceExceptions and raises a
        # owslib.wms.ServiceException. However, openURL only checks for mime
        # types text/xml and application/xml. application/vnd.ogc.se_xml is
        # not considered. For some reason, the check below doesn't work, though..
        proxies = config_loader(dataset="proxies")

        u = ogcwms.openURL(base_url, data, method,
                           username=self.auth.username, password=self.auth.password, proxies=proxies)

        # check for service exceptions, and return
        # NOTE: There is little bug in owslib.util.openURL -- if the file
        # returned by the http server is an XML file, urlopen converts it
        # into an "RereadableURL" object. WHile this enables the URL content
        # to be scanned in urlopen as well as in a following method
        # (urllib2.urlopen objects only allow the content to be read once),
        # the "info" attribute is missing after the conversion..
        if hasattr(u, "info") and 'application/vnd.ogc.se_xml' in u.info()['Content-Type']:
            se_xml = u.read()
            se_tree = etree.fromstring(se_xml)
            # Remove namespaces in the response, otherwise this code might fail
            # (mslib) add ogcwms
            ogcwms.removeXMLNamespace(se_tree)
            err_message = str(se_tree.find('ServiceException').text).strip()
            raise owslib.util.ServiceException(err_message, se_xml)
        return u

    def get_redirect_url(self, method="Get"):
        # return self.getOperationByName("GetMap").methods[method]["url"]
        return self.getOperationByName("GetMap").methods[0]["url"]


class MSS_WMS_AuthenticationDialog(QtWidgets.QDialog, ui_pw.Ui_WMSAuthenticationDialog):
    """Dialog to ask the user for username/password should this be
       required by a WMS server.
    """

    def __init__(self, parent=None):
        """
        Arguments:
        parent -- Qt widget that is parent to this widget.
        """
        super(MSS_WMS_AuthenticationDialog, self).__init__(parent)
        self.setupUi(self)

    def getAuthInfo(self):
        """Return the entered username and password.
        """
        return (self.leUsername.text(),
                self.lePassword.text())


class WMSMapFetcher(QtCore.QObject):
    """
    This class is supposed to run in a background thread to prefetch map images that may be used later on.

    This class uses code that is very similar to the actual code in the getMap call. Might be refactored
    in some way...
    """

    # start fetching next URL in queue
    process = QtCore.pyqtSignal()
    # present a final image including legend
    finished = QtCore.pyqtSignal([object, object, object, object, object, object, object])
    # triggered in case of caught exception
    exception = QtCore.pyqtSignal([Exception])
    # triggered in case of long lasting operation
    started_request = QtCore.pyqtSignal()

    def __init__(self, wms_cache, parent=None):
        super(WMSMapFetcher, self).__init__(parent)
        self.wms_cache = wms_cache
        self.maps = []
        self.map_imgs = []
        self.legend_imgs = []
        self.process.connect(self.process_map, QtCore.Qt.QueuedConnection)
        self.long_request = False

    @QtCore.pyqtSlot(list)
    def fetch_maps(self, map_list):
        """
        Initializes the list of maps to be fetched. Overwrites any remaining ones.
        Starts processing loop
        """
        self.maps = map_list
        self.process.emit()

    @QtCore.pyqtSlot()
    def process_map(self):
        """
        Processing. Self emits into event queue until list is empty. In this way,
        fetch_map events may interrupt.
        """
        if len(self.maps) == 0:
            return
        layer, kwargs, md5_filename, use_cache, legend_kwargs = self.maps[0]
        self.maps = self.maps[1:]
        self.long_request = False
        try:
            self.map_imgs.append(self.fetch_map(layer, kwargs, use_cache, md5_filename))
            self.legend_imgs.append(self.fetch_legend(use_cache=use_cache, **legend_kwargs))
        except Exception as ex:
            logging.error("MapPrefetcher Exception %s - %s.", type(ex), ex)
            # emit finished so progress dialog will be closed
            self.finished.emit(None, None, None, None, None, None, md5_filename)
            self.exception.emit(ex)
            self.map_imgs = []
            self.legend_imgs = []
        else:
            if len(self.maps) > 0:
                self.process.emit()
            else:
                self.finished.emit(
                    self.map_imgs, self.legend_imgs, kwargs["layers"][0], kwargs["styles"][0], kwargs["init_time"],
                    kwargs["time"], md5_filename)
                self.map_imgs = []
                self.legend_imgs = []

    def fetch_map(self, layer, kwargs, use_cache, md5_filename):
        """
        Retrieves a WMS map from a server or reads it from a cache if allowed
        """
        logging.debug("MapPrefetcher %s %s.", kwargs["time"], kwargs["level"])

        if use_cache and os.path.exists(md5_filename):
            if ".png" in md5_filename:
                img = Image.open(md5_filename)
                img.load()
                logging.debug("MapPrefetcher - found image cache")
            else:
                with open(md5_filename, "r") as cache:
                    return etree.fromstring(cache.read())
        else:
            self.started_request.emit()
            self.long_request = True
            urlobject = layer.get_wms().getmap(**kwargs)

            if "xml" in urlobject.info()["content-type"].lower():
                with open(md5_filename, "w") as cache:
                    cache.write(str(urlobject.read(), encoding="utf8"))
                return etree.fromstring(urlobject.read())

            image_io = io.BytesIO(urlobject.read())
            img = Image.open(image_io)
            # Check if the image is stored as indexed palette
            # with a transparent colour. Store correspondingly.
            if img.mode == "P" and "transparency" in img.info:
                img.save(md5_filename, transparency=img.info["transparency"])
            else:
                img.save(md5_filename)
            logging.debug("MapPrefetcher %s - saved filed: %s.", self, md5_filename)
        return img.convert("RGBA")

    def fetch_legend(self, urlstr=None, use_cache=True, md5_filename=None):
        """
        Retrieves a WMS legend from a server or reads it from a cache if allowed
        """
        if urlstr is None:
            return None

        if use_cache and os.path.exists(md5_filename):
            legend_img = Image.open(md5_filename)
            legend_img.load()
            logging.debug("MapPrefetcher - found legend cache")
        else:
            if not self.long_request:
                self.started_request.emit()
                self.long_request = True
            # This StringIO object can then be passed as a file substitute to
            # Image.open(). See
            #    http://www.pythonware.com/library/pil/handbook/image.htm
            logging.debug("Retrieving legend from '%s'", urlstr)
            urlobject = requests.get(urlstr)
            image_io = io.BytesIO(urlobject.content)
            try:
                legend_img_raw = Image.open(image_io)
            except Exception as ex:
                # This exception may be triggered if there was a problem with the legend
                # as present with http://geoservices.knmi.nl/cgi-bin/HARM_N25.cgi
                # it is deemed preferential to display the WMS map and forget about the
                # legend than not displaying anything.
                logging.error("Wildecard Exception %s - %s.", type(ex), ex)
                return None
            legend_img = legend_img_raw.crop(legend_img_raw.getbbox())
            # Store the retrieved image in the cache, if enabled.
            try:
                legend_img.save(md5_filename, transparency=0)
            except Exception as ex:
                logging.debug("Wildecard Exception %s - %s.", type(ex), ex)
                legend_img.save(md5_filename)
        return legend_img.convert("RGBA")


class WMSControlWidget(QtWidgets.QWidget, ui.Ui_WMSDockWidget):
    """The base class of the WMS control widget: Provides the GUI
       elements and common functionality to access a WMS. This class
       is not instantiated directly, see HSecWMSControlWidget and
       VSecWMSControlWidget below.

    NOTE: Currently this class can only handle WMS 1.1.1/1.3.0 servers.
    """

    prefetch = QtCore.pyqtSignal([list], name="prefetch")
    fetch = QtCore.pyqtSignal([list], name="fetch")
    signal_disable_cbs = QtCore.Signal(name="disable_cbs")
    signal_enable_cbs = QtCore.Signal(name="enable_cbs")
    image_displayed = QtCore.pyqtSignal()

    def __init__(self, parent=None, default_WMS=None, wms_cache=None, view=None):
        """
        Arguments:
        parent -- Qt widget that is parent to this widget.
        default_WMS -- list of strings that specify WMS URLs that will be
                       displayed in the URL combobox as default values.
        """
        super(WMSControlWidget, self).__init__(parent)
        self.setupUi(self)

        self.view = view

        # Multilayering things
        self.multilayers = Multilayers(self)
        self.pbLayerSelect.clicked.connect(lambda: (self.multilayers.hide(), self.multilayers.show()))
        self.multilayers.cbWMS_URL.editTextChanged.connect(
            lambda text: self.multilayers.pbViewCapabilities.setEnabled(text in self.multilayers.layers))

        # Initial list of WMS servers.
        self.multilayers.cbWMS_URL.setModel(WMS_URL_LIST)
        if default_WMS is not None:
            add_wms_urls(self.multilayers.cbWMS_URL, default_WMS)
        # set last connected url to editable
        wms_settings = load_settings_qsettings('wms', {'recent_wms_url': None})
        if wms_settings['recent_wms_url'] is not None:
            add_wms_urls(self.multilayers.cbWMS_URL, [wms_settings['recent_wms_url']])

        self.layerChangeInProgress = False
        self.save_level = None

        # Initialise GUI elements that control WMS parameters.
        self.cbLevel.clear()
        self.cbInitTime.clear()
        self.cbValidTime.clear()

        self.time_steps = ["1 min", "5 min", "10 min", "15 min", "30 min",
                           "1 hour", "2 hours", "3 hours", "6 hours",
                           "12 hours", "24 hours", "2 days", "7 days"]
        self.cbInitTime_step.clear()
        self.cbInitTime_step.addItems(self.time_steps)
        self.cbInitTime_step.setCurrentIndex(self.cbInitTime_step.findText("12 hours"))
        self.cbValidTime_step.clear()
        self.cbValidTime_step.addItems(self.time_steps)
        self.cbValidTime_step.setCurrentIndex(self.cbInitTime_step.findText("3 hours"))

        self.enable_level_elements(False)
        self.enable_valid_time_elements(False)
        self.enable_init_time_elements(False)
        self.btGetMap.setEnabled(False)
        self.multilayers.pbViewCapabilities.setEnabled(False)

        self.cbTransparent.setChecked(False)

        # Check for WMS image cache directory, create if necessary.
        if wms_cache is not None:
            self.wms_cache = os.path.join(wms_cache, "")
            logging.debug("checking for WMS image cache at %s ...", self.wms_cache)
            if not os.path.exists(self.wms_cache):
                logging.debug("  created new image cache directory.")
                os.makedirs(self.wms_cache)
            else:
                logging.debug("  found.")
            # Service the cache (delete files that are too old, remove oldest
            # files if cache is too large).
            self.service_cache()
        else:
            self.wms_cache = None

        # Initialise date/time fields with current day, 00 UTC.
        self.dteInitTime.setDateTime(QtCore.QDateTime(
            datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)))
        self.dteValidTime.setDateTime(QtCore.QDateTime(
            datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)))

        # Connect slots and signals.
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        self.multilayers.btGetCapabilities.clicked.connect(self.get_capabilities)
        self.multilayers.pbViewCapabilities.clicked.connect(self.view_capabilities)

        self.btClearMap.clicked.connect(self.clear_map)

        self.cbLevel.currentIndexChanged.connect(self.level_changed)
        self.cbInitTime.currentIndexChanged.connect(self.init_time_changed)
        self.cbValidTime.currentIndexChanged.connect(self.valid_time_changed)
        self.multilayers.needs_repopulate.connect(self.populate_ui)

        self.tbInitTime_back.clicked.connect(self.init_time_back_click)
        self.tbInitTime_fwd.clicked.connect(self.init_time_fwd_click)
        self.tbInitTime_cbback.clicked.connect(self.cb_init_time_back_click)
        self.tbInitTime_cbfwd.clicked.connect(self.cb_init_time_fwd_click)
        self.dteInitTime.dateTimeChanged.connect(self.check_init_time)
        self.tbValidTime_back.clicked.connect(self.valid_time_back_click)
        self.tbValidTime_fwd.clicked.connect(self.valid_time_fwd_click)
        self.tbValidTime_cbback.clicked.connect(self.cb_valid_time_back_click)
        self.tbValidTime_cbfwd.clicked.connect(self.cb_valid_time_fwd_click)
        self.dteValidTime.dateTimeChanged.connect(self.check_valid_time)
        self.tbLevel_back.clicked.connect(self.level_back_click)
        self.tbLevel_fwd.clicked.connect(self.level_fwd_click)

        self.btClearCache.clicked.connect(self.clearCache)
        self.multilayers.cbWMS_URL.editTextChanged.connect(self.wms_url_changed)
        if view is not None and hasattr(view, "redrawn"):
            self.view.redrawn.connect(self.after_redraw)

        # Progress dialog to inform the user about image ongoing retrievals.
        self.pdlg = QtWidgets.QProgressDialog(
            "retrieving image...", "Cancel", 0, 10, parent=self.parent())
        self.pdlg.close()

        # Progress dialog to inform the user about ongoing capability requests.
        self.capabilities_worker = Worker(None)
        self.cpdlg = QtWidgets.QProgressDialog(
            "retrieving wms capabilities...", "Cancel", 0, 10, parent=self.multilayers)
        self.cpdlg.canceled.connect(self.stop_capabilities_retrieval)
        self.cpdlg.close()

        self.thread_prefetch = QtCore.QThread()  # no parent!
        self.thread_prefetch.start()
        self.thread_fetch = QtCore.QThread()  # no parent!
        self.thread_fetch.start()

        self.prefetcher = None
        self.fetcher = None
        self.expected_img = None
        self.check_for_allowed_crs = True

        if self.multilayers.cbWMS_URL.count() > 0:
            self.multilayers.cbWMS_URL.setCurrentIndex(0)
            self.wms_url_changed(self.multilayers.cbWMS_URL.currentText())

    def __del__(self):
        """Destructor.
        """
        # Service the WMS cache on application exit.
        if self.wms_cache is not None:
            self.service_cache()
        # properly terminate background threads. wait is necessary!
        self.thread_prefetch.quit()
        self.thread_prefetch.wait()
        self.thread_fetch.quit()
        self.thread_fetch.wait()

    def get_all_maps(self, disregard_current=False):
        if self.multilayers.cbMultilayering.isChecked():
            self.get_map(self.multilayers.get_active_layers())
        else:
            self.get_map([self.multilayers.get_current_layer()])

    def initialise_wms(self, base_url, version="1.3.0"):
        """Initialises a MSSWebMapService object with the specified base_url.

        If the web server returns a '401 Unauthorized', prompt the user for
        username and password.

        NOTE: owslib (from which MSSWebMapService is derived) only supports
        the basic HTTP authentication. You might hence have to dig deeper into
        the code for more sophisticated authentication methods.

        See: 1) http://pythonpaste.org/modules/auth.basic.html#module-paste.auth.basic
             2) method 'openURL' in owslib.util

        (mr, 2011-02-25)
        """
        # initialize login cache fomr config file, but do not overwrite existing keys
        for key, value in config_loader(dataset="WMS_login").items():
            if key not in constants.WMS_LOGIN_CACHE:
                constants.WMS_LOGIN_CACHE[key] = value
        username, password = constants.WMS_LOGIN_CACHE.get(base_url, (None, None))

        def on_success(wms):
            self.cpdlg.setValue(9)
            if wms is not None:

                # update the combo box, if entry requires change/insertion
                found = False
                current_url = self.multilayers.cbWMS_URL.currentText()
                for count in range(self.multilayers.cbWMS_URL.count()):
                    if self.multilayers.cbWMS_URL.itemText(count) == current_url:
                        self.multilayers.cbWMS_URL.setItemText(count, base_url)
                        self.multilayers.cbWMS_URL.setCurrentIndex(count)
                        found = True
                        break
                    if self.multilayers.cbWMS_URL.itemText(count) == base_url:
                        self.multilayers.cbWMS_URL.setCurrentIndex(count)
                        found = True
                if not found:
                    logging.debug("inserting URL: %s", base_url)
                    add_wms_urls(self.multilayers.cbWMS_URL, [base_url])
                    self.multilayers.cbWMS_URL.setEditText(base_url)
                    save_settings_qsettings('wms', {'recent_wms_url': base_url})

                self.activate_wms(wms)
                WMS_SERVICE_CACHE[wms.url] = wms
                self.cpdlg.close()

        def on_failure(e):
            try:
                raise e
            except owslib.util.ServiceException as ex:
                logging.error("ERROR: %s %s", type(ex), ex)
                if str(ex).startswith("401") or str(ex).find("Error 401") >= 0 or str(ex).find(
                        "Unauthorized") >= 0:
                    # Catch the "401 Unauthorized" error if one has been
                    # returned by the server and ask the user for username
                    # and password.
                    # WORKAROUND (mr, 28Mar2013) -- owslib doesn't recognize
                    # the Apache 401 messages, we get an XML message here but
                    # no error code. Quick workaround: Scan the XML message for
                    # the string "Error 401"...
                    dlg = MSS_WMS_AuthenticationDialog(parent=self.multilayers)
                    dlg.setModal(True)
                    if dlg.exec_() == QtWidgets.QDialog.Accepted:
                        username, password = dlg.getAuthInfo()
                        # If user & pw have been entered, cache them.
                        constants.WMS_LOGIN_CACHE[base_url] = (username, password)
                        self.capabilities_worker.function = lambda: MSSWebMapService(
                            base_url, version=version,
                            username=username, password=password)
                        self.capabilities_worker.start()
                    else:
                        self.cpdlg.close()
                        return
                else:
                    logging.error("cannot load capabilities document.. "
                                  "no layers can be used in this view.")
                    QtWidgets.QMessageBox.critical(
                        self.multilayers, self.tr("Web Map Service"),
                        self.tr(f"ERROR: We cannot load the capability document!\n\n{type(ex)}\n{ex}"))
                    self.cpdlg.close()
            except Exception as ex:
                logging.error("cannot load capabilities document.. "
                              "no layers can be used in this view.")
                QtWidgets.QMessageBox.critical(
                    self.multilayers, self.tr("Web Map Service"),
                    self.tr(f"ERROR: We cannot load the capability document!\n\n{type(ex)}\n{ex}"))
                self.cpdlg.close()

        try:
            str(base_url)
        except UnicodeEncodeError:
            logging.error("got a unicode url?!: '%s'", base_url)
            QtWidgets.QMessageBox.critical(self.multilayers, self.tr("Web Map Service"),
                                           self.tr("ERROR: We cannot parse unicode URLs!"))
            self.cpdlg.close()

        Worker.create(lambda: MSSWebMapService(base_url, version=version, username=username, password=password),
                      on_success, on_failure)

    def wms_url_changed(self, text):
        wms = WMS_SERVICE_CACHE.get(text)
        if wms is not None:
            self.activate_wms(wms, cache=True)

    @QtCore.pyqtSlot(Exception)
    def display_exception(self, ex):
        logging.error("ERROR: %s %s", type(ex), ex)
        logging.debug("%s", traceback.format_exc())
        QtWidgets.QMessageBox.critical(
            self, self.tr("Web Map Service"), self.tr(f"ERROR:\n{type(ex)}\n{ex}"))

    @QtCore.pyqtSlot()
    def display_progress_dialog(self):
        logging.debug("showing progress dialog")
        self.pdlg.reset()
        self.pdlg.setValue(5)
        self.pdlg.setModal(True)
        self.pdlg.show()

    def display_capabilities_dialog(self):
        logging.debug("showing capabilities dialog")
        self.cpdlg.reset()
        self.cpdlg.setValue(1)
        self.cpdlg.setModal(True)
        self.cpdlg.show()

    def stop_capabilities_retrieval(self):
        logging.debug("stopping capabilities retrieval")
        try:
            self.capabilities_worker.quit()
            self.capabilities_worker.disconnect()
        except TypeError:
            pass

    def clear_map(self):
        logging.debug("clear figure")
        self.view.clear_figure()
        logging.debug("enabling checkboxes in map-options if any")
        self.signal_enable_cbs.emit()

    def get_capabilities(self):
        """
        Query the WMS server in the URL combobox for its capabilities.
        """

        # Load new WMS. Only add those layers to the combobox that can provide
        # the CRS that match the filter of this module.
        base_url = self.multilayers.cbWMS_URL.currentText()
        params = {'service': 'WMS',
                  'request': 'GetCapabilities'}

        def on_success(request):
            self.cpdlg.setValue(5)
            # url shortener url translated
            url = request.url

            url = url.replace("?service=WMS", "").replace("&service=WMS", "") \
                .replace("?request=GetCapabilities", "").replace("&request=GetCapabilities", "")
            logging.debug("requesting capabilities from %s", url)
            self.initialise_wms(url, None)

        def on_failure(e):
            try:
                raise e
            except (requests.exceptions.TooManyRedirects,
                    requests.exceptions.ConnectionError,
                    requests.exceptions.InvalidURL,
                    requests.exceptions.InvalidSchema,
                    requests.exceptions.MissingSchema) as ex:
                logging.error("Cannot load capabilities document.\n"
                              "No layers can be used in this view.")
                QtWidgets.QMessageBox.critical(
                    self.multilayers, self.tr("Web Map Service"),
                    self.tr(f"ERROR: We cannot load the capability document!\n\\n{type(ex)}\n{ex}"))
            finally:
                self.cpdlg.close()

        self.display_capabilities_dialog()
        Worker.create(lambda: requests.get(base_url, params=params),
                      on_success, on_failure)

    def activate_wms(self, wms, cache=False):
        # Parse layer tree of the wms object and discover usable layers.
        stack = list(wms.contents.values())
        filtered_layers = set()
        while len(stack) > 0:
            layer = stack.pop()
            if len(layer.layers) > 0:
                stack.extend(layer.layers)
            elif self.is_layer_aligned(layer):
                cb_string = f"{layer.title} | {layer.name}"
                filtered_layers.add(cb_string)
        logging.debug("discovered %i layers that can be used in this view",
                      len(filtered_layers))
        filtered_layers = sorted(filtered_layers)
        selected = self.multilayers.current_layer.text(0) if self.multilayers.current_layer else None
        if not cache and wms.url in self.multilayers.layers and \
                wms.capabilities_document.decode("utf-8") != \
                self.multilayers.layers[wms.url]["wms"].capabilities_document.decode("utf-8"):
            self.multilayers.delete_server(self.multilayers.layers[wms.url]["header"])
        self.multilayers.add_wms(wms)
        for layer in filtered_layers:
            self.multilayers.add_multilayer(layer, wms, layer == selected)
        self.multilayers.filter_multilayers()
        self.multilayers.update_checkboxes()
        self.multilayers.pbViewCapabilities.setEnabled(True)
        if len(filtered_layers) > 0:
            self.populate_ui()

        if self.prefetcher is not None:
            self.prefetch.disconnect(self.prefetcher.fetch_maps)
        if self.fetcher is not None:
            self.fetch.disconnect(self.fetcher.fetch_maps)

        self.prefetcher = WMSMapFetcher(self.wms_cache)
        self.prefetcher.moveToThread(self.thread_prefetch)
        self.prefetch.connect(self.prefetcher.fetch_maps)  # implicitely uses a queued connection

        self.fetcher = WMSMapFetcher(self.wms_cache)
        self.fetcher.moveToThread(self.thread_fetch)
        self.fetch.connect(self.fetcher.fetch_maps)  # implicitely uses a queued connection
        self.fetcher.finished.connect(self.continue_retrieve_image)  # implicitely uses a queued connection
        self.fetcher.exception.connect(self.display_exception)  # implicitely uses a queued connection
        self.fetcher.started_request.connect(self.display_progress_dialog)  # implicitely uses a queued connection

        # logic to disable fill continents, fill oceans on connection to
        self.signal_disable_cbs.emit()

    def view_capabilities(self):
        """Open a WMSCapabilitiesBrowser dialog showing the capabilities
           document.
        """
        logging.debug("Opening WMS capabilities browser..")
        wms = self.multilayers.layers[self.multilayers.cbWMS_URL.currentText()]["wms"]
        if wms is not None:
            wmsbrws = wms_capabilities.WMSCapabilitiesBrowser(
                parent=self.multilayers,
                url=wms.url,
                capabilities=wms)
            wmsbrws.setAttribute(QtCore.Qt.WA_DeleteOnClose)
            wmsbrws.show()

    def get_layer_object(self, wms, layername):
        """Returns the object from the layer tree that fits the given
           layer name.
        """
        if wms is None:
            return None
        if layername in wms.contents:
            return wms.contents[layername]
        else:
            stack = list(wms.contents.values())
            while len(stack) > 0:
                layer = stack.pop()
                if layer.name == layername:
                    return layer
                if len(layer.layers) > 0:
                    stack.extend(layer.layers)
        return None

    def parse_time_extent(self, values):
        times = []
        for time_item in [i.strip() for i in values]:
            try:
                list_desc = time_item.split("/")

                if len(list_desc) == 3:
                    time_val = parse_iso_datetime(list_desc[0])
                    if "current" in list_desc[1]:
                        end_time = datetime.utcnow()
                    else:
                        end_time = parse_iso_datetime(list_desc[1])
                    delta = parse_iso_duration(list_desc[2])
                    while time_val <= end_time:
                        times.append(time_val)
                        time_val += delta

                elif len(list_desc) == 1:
                    times.append(parse_iso_datetime(time_item))
                else:
                    raise ValueError("value has incorrect number of entries.")

            except Exception as ex:
                logging.debug("Wildcard Exception %s - %s.", type(ex), ex)
                logging.error("Can't understand time string '%s'. Please check the implementation.", time_item)
        return times

    def disable_ui(self):
        self.disable_cbInitTime_elements()
        self.disable_cbValidTime_elements()
        self.disable_dteInitTime_elements()
        self.disable_dteValidTime_elements()
        self.enable_valid_time_elements(False)
        self.enable_init_time_elements(False)
        self.enable_level_elements(False)
        self.btGetMap.setEnabled(False)

    def populate_ui(self):
        """
        Adds the values of the current layer to the UI comboboxes
        """
        self.layerChangeInProgress = True
        self.cbLevel.clear()
        self.cbInitTime.clear()
        self.cbValidTime.clear()

        active_layers = self.multilayers.get_active_layers()
        layer = self.multilayers.get_current_layer()

        if not layer:
            self.lLayerName.setText("No Layer selected")
            self.disable_ui()
            return

        else:
            self.btGetMap.setEnabled(True)

        self.lLayerName.setText(str(layer))

        tooltip_text = ""
        for active_layer in active_layers if layer.checkState(0) else [layer]:
            tooltip_text += f"{active_layer}\n"
        self.lLayerName.setToolTip(tooltip_text.strip())

        if len(active_layers) > 1 and layer.checkState(0):
            self.lLayerName.setText(self.lLayerName.text() + f" (and {len(active_layers) - 1} more)")

        crs = layer.get_allowed_crs()
        levels, itimes, vtimes = layer.get_levels(), layer.get_itimes(), layer.get_vtimes()
        if vtimes and itimes:
            vtimes = vtimes[next((i for i, vtime in enumerate(vtimes) if vtime >= layer.get_itime()), 0):]

        if levels:
            self.cbLevel.addItems(levels)
            self.cbLevel.setCurrentIndex(self.cbLevel.findText(layer.get_level()))
        self.enable_level_elements(len(levels) > 0)

        if itimes:
            self.cbInitTime.addItems(itimes)
            self.cbInitTime.setCurrentIndex(self.cbInitTime.findText(layer.get_itime()))
        self.enable_init_time_elements(len(itimes) > 0)

        if vtimes:
            self.cbValidTime.addItems(vtimes)
            self.cbValidTime.setCurrentIndex(self.cbValidTime.findText(layer.get_vtime()))
        self.enable_valid_time_elements(len(vtimes) > 0)

        if not self.init_time_changed():
            self.disable_dteInitTime_elements()
        if not self.valid_time_changed():
            self.disable_dteValidTime_elements()
        if self.cbInitTime.count() == 0:
            self.disable_cbInitTime_elements()
        if self.cbValidTime.count() == 0:
            self.disable_cbValidTime_elements()

        if crs and \
           self.parent() is not None and \
           self.parent().parent() is not None:
            logging.debug("Layer offers '%s' projections.", crs)
            extra = [_code for _code in crs if _code.startswith("EPSG")]
            extra = [_code for _code in sorted(extra) if _code[5:] in basemap.epsg_dict]
            logging.debug("Selected '%s' for Combobox.", extra)
            self.parent().parent().update_predefined_maps(extra)

        self.layerChangeInProgress = False

    @staticmethod
    def secs_from_timestep(timestep_string):
        """Convert a string specifying a time step (e.g. 5 min, 3 hours) to
           seconds.

        Allowed formats:
         a) 'XX min', where XX is an integer.
         b) 'XX hours', where XX is an integer.
         c) 'XX days', where XX is an integer.
        """
        try:
            return 60 * int(timestep_string.split(" min")[0])
        except ValueError as error:
            logging.debug("ValueError Exception %s", error)
        try:
            return 3600 * int(timestep_string.split(" hour")[0])
        except ValueError as error:
            logging.debug("ValueError Exception %s", error)
        try:
            return 86400 * int(timestep_string.split(" days")[0])
        except ValueError as error:
            logging.debug("ValueError Exception %s", error)
            raise ValueError(f"cannot convert '{timestep_string}' to seconds: wrong format.")

    def init_time_back_click(self):
        """Slot for the tbInitTime_back button.
        """
        # Get QDateTime object from QtDateTimeEdit field.
        d = self.dteInitTime.dateTime()
        # Add value from cbInitTime_step and set new date.
        secs = self.secs_from_timestep(self.cbInitTime_step.currentText())
        self.dteInitTime.setDateTime(d.addSecs(-secs))
        self.auto_update()

    def init_time_fwd_click(self):
        """Slot for the tbInitTime_fwd button.
        """
        # Get QDateTime object from QtDateTimeEdit field.
        d = self.dteInitTime.dateTime()
        # Add value from cbInitTime_step and set new date.
        secs = self.secs_from_timestep(self.cbInitTime_step.currentText())
        self.dteInitTime.setDateTime(d.addSecs(secs))
        self.auto_update()

    def valid_time_back_click(self):
        # Get QDateTime object from QtDateTimeEdit field.
        d = self.dteValidTime.dateTime()
        # Add value from cbInitTime_step and set new date.
        secs = self.secs_from_timestep(self.cbValidTime_step.currentText())
        self.dteValidTime.setDateTime(d.addSecs(-secs))
        self.auto_update()

    def valid_time_fwd_click(self):
        # Get QDateTime object from QtDateTimeEdit field.
        d = self.dteValidTime.dateTime()
        # Add value from cbInitTime_step and set new date.
        secs = self.secs_from_timestep(self.cbValidTime_step.currentText())
        self.dteValidTime.setDateTime(d.addSecs(secs))
        self.auto_update()

    def level_back_click(self):
        ci = self.cbLevel.currentIndex()
        if ci > 0:
            ci = ci - 1
        self.cbLevel.setCurrentIndex(ci)

    def level_fwd_click(self):
        ci = self.cbLevel.currentIndex()
        if ci < self.cbLevel.count() - 1:
            ci = ci + 1
        self.cbLevel.setCurrentIndex(ci)

    def cb_init_time_back_click(self):
        ci = self.cbInitTime.currentIndex()
        if ci > 0:
            ci = ci - 1
        self.cbInitTime.setCurrentIndex(ci)
        self.init_time_changed()

    def cb_init_time_fwd_click(self):
        ci = self.cbInitTime.currentIndex()
        if ci < self.cbInitTime.count() - 1:
            ci = ci + 1
        self.cbInitTime.setCurrentIndex(ci)
        self.init_time_changed()

    def cb_valid_time_back_click(self):
        ci = self.cbValidTime.currentIndex()
        if ci > 0:
            ci = ci - 1
        self.cbValidTime.setCurrentIndex(ci)
        self.valid_time_changed()

    def cb_valid_time_fwd_click(self):
        ci = self.cbValidTime.currentIndex()
        if ci < self.cbValidTime.count() - 1:
            ci = ci + 1
        self.cbValidTime.setCurrentIndex(ci)
        self.valid_time_changed()

    def auto_update(self):
        """If the auto update check box is checked, let btGetMap emit a
           clicked() signal everytime this method is called.
           autoUpdate() should be called from the slots that handle
           time and level changes.

        If the the layerChangeInProgress flag is set no update will be
        performed (lock mechanism to prevent erroneous requests during
        a layer change).
        """
        if self.btGetMap.isEnabled() and self.cbAutoUpdate.isChecked() and not self.layerChangeInProgress:
            self.get_all_maps()

    def check_init_time(self, dt):
        """Checks whether an initialisation time set with the init time
           date/time edit is contained in the list of init times advertised
           by the WMS server.

        If the time is not contained in the list, the font will be set to
        'strike through' to indicate an invalid time.
        """
        font = self.dteInitTime.font()
        pydt = dt.toPyDateTime()
        init_time_available = pydt in self.multilayers.get_current_layer().allowed_init_times
        font.setStrikeOut(not init_time_available)
        self.dteInitTime.setFont(font)
        if init_time_available:
            index = self.cbInitTime.findText(pydt.isoformat() + "Z")
            self.cbInitTime.setCurrentIndex(index)

    def check_valid_time(self, dt):
        """Same as check_init_time, but for valid time.
        """
        valid_time_available = True
        pydt = dt.toPyDateTime()
        if self.multilayers.get_current_layer().allowed_valid_times:
            if pydt in self.multilayers.get_current_layer().allowed_valid_times:
                index = self.cbValidTime.findText(pydt.isoformat() + "Z")
                # setCurrentIndex also sets the date/time edit via signal.
                if index > -1:
                    self.cbValidTime.setCurrentIndex(index)
                else:
                    valid_time_available = False
            else:
                valid_time_available = False
        font = self.dteValidTime.font()
        font.setStrikeOut(not valid_time_available)
        self.dteValidTime.setFont(font)

    def init_time_changed(self):
        """Slot to be called when the current index of the init time
           combo box is changed. The method tries to sync to the
           init time date/time edit.
        """
        init_time = self.cbInitTime.currentText()
        if init_time != "":
            init_time = parse_iso_datetime(init_time)
            if init_time is not None:
                self.dteInitTime.setDateTime(init_time)

        if self.multilayers.threads == 0 and not self.layerChangeInProgress:
            self.multilayers.get_current_layer().set_itime(self.cbInitTime.currentText())
            self.multilayers.carry_parameters["itime"] = self.cbInitTime.currentText()

        self.auto_update()
        return init_time == "" or init_time is not None

    def valid_time_changed(self):
        """Same as initTimeChanged(), but for the valid time elements.
        """
        valid_time = self.cbValidTime.currentText()
        if valid_time != "":
            valid_time = parse_iso_datetime(valid_time)
            if valid_time is not None:
                self.dteValidTime.setDateTime(valid_time)

        if self.multilayers.threads == 0 and not self.layerChangeInProgress:
            self.multilayers.get_current_layer().set_vtime(self.cbValidTime.currentText())
            self.multilayers.carry_parameters["vtime"] = self.cbValidTime.currentText()

        self.auto_update()
        return valid_time == "" or valid_time is not None

    def level_changed(self):
        if self.multilayers.threads == 0 and not self.layerChangeInProgress:
            self.multilayers.get_current_layer().set_level(self.cbLevel.currentText())
            self.multilayers.carry_parameters["level"] = self.cbLevel.currentText()
        self.auto_update()

    def enable_level_elements(self, enable):
        """Enable or disable the GUI elements allowing vertical elevation
           level control.
        """
        self.cbLevel.setEnabled(enable and self.cbLevel.count() > 1)
        self.tbLevel_back.setEnabled(enable)
        self.tbLevel_fwd.setEnabled(enable)

    def enable_init_time_elements(self, enable):
        """Enables or disables the GUI elements allowing initialisation time
           control.
        """
        self.cbInitTime.setEnabled(enable)
        self.dteInitTime.setEnabled(enable)
        self.tbInitTime_back.setEnabled(enable)
        self.tbInitTime_fwd.setEnabled(enable)
        self.tbInitTime_cbback.setEnabled(enable)
        self.tbInitTime_cbfwd.setEnabled(enable)
        self.cbInitTime_step.setEnabled(enable)

    def enable_valid_time_elements(self, enable):
        """Enables or disables the GUI elements allowing valid time
           control.
        """
        self.cbValidTime.setEnabled(enable)
        self.dteValidTime.setEnabled(enable)
        self.tbValidTime_back.setEnabled(enable)
        self.tbValidTime_fwd.setEnabled(enable)
        self.tbValidTime_cbback.setEnabled(enable)
        self.tbValidTime_cbfwd.setEnabled(enable)
        self.cbValidTime_step.setEnabled(enable)

    def disable_dteInitTime_elements(self):
        """Disables init time date/time edit elements.
        """
        enable = False
        self.dteInitTime.setEnabled(enable)
        self.tbInitTime_back.setEnabled(enable)
        self.tbInitTime_fwd.setEnabled(enable)
        self.cbInitTime_step.setEnabled(enable)

    def disable_dteValidTime_elements(self):
        """Disables valid time date/time edit elements.
        """
        enable = False
        self.dteValidTime.setEnabled(enable)
        self.tbValidTime_back.setEnabled(enable)
        self.tbValidTime_fwd.setEnabled(enable)
        self.cbValidTime_step.setEnabled(enable)

    def disable_cbInitTime_elements(self):
        """Disables init time combobox elements.
        """
        enable = False
        self.cbInitTime.setEnabled(enable)
        self.tbInitTime_cbback.setEnabled(enable)
        self.tbInitTime_cbfwd.setEnabled(enable)

    def disable_cbValidTime_elements(self):
        """Disables valid time combobox elements.
        """
        enable = False
        self.cbValidTime.setEnabled(enable)
        self.tbValidTime_cbback.setEnabled(enable)
        self.tbValidTime_cbfwd.setEnabled(enable)

    def get_layer(self):
        return self.multilayers.get_current_layer().get_layer()

    def get_style(self):
        return self.multilayers.get_current_layer().get_style()

    def get_level(self):
        return self.multilayers.get_current_layer().get_level_name()

    def get_init_time(self):
        """Get the initialisation time from the GUI elements.

        If the init time date/time edit is enabled (i.e. the times specified
        in the WMS capabilities document can be interpreted), return a
        datetime object of the currently set time. Otherwise, return the
        string that is selected in the init time combobox.
        """
        if self.cbInitialisationOn.isChecked():
            if self.dteInitTime.isEnabled():
                return self.dteInitTime.dateTime().toPyDateTime()
            else:
                itime_str = self.multilayers.get_current_layer().get_itime()
                return itime_str
        else:
            return None

    def get_valid_time(self):
        """The same as getInitTime(), but for the valid time.
        """
        if self.cbValidOn.isChecked():
            if self.dteValidTime.isEnabled():
                return self.dteValidTime.dateTime().toPyDateTime()
            else:
                vtime_str = self.multilayers.get_current_layer().get_vtime()
                return vtime_str
        else:
            return None

    def caching_enabled(self):
        """Returns if the image cache is enabled.
        """
        return self.wms_cache is not None and self.cbCacheEnabled.isChecked()

    def get_md5_filename(self, layer, kwargs):
        urlstr = layer.get_wms().getmap(return_only_url=True, **kwargs)
        ending = ".png" if "image" in kwargs["format"] else ".xml"
        return os.path.join(self.wms_cache, hashlib.md5(urlstr.encode('utf-8')).hexdigest() + ending)

    def retrieve_image(self, layers=None, crs="EPSG:4326", bbox=None, path_string=None,
                       width=800, height=400, transparent=False, format="image/png"):
        """Retrieve an image of the layer currently selected in the
           GUI elements from the current WMS provider. If caching is
           enabled, first check the cache for the requested image. If
           the image is retrieved from the WMS and caching is enabled,
           store the image into the cache.
           If a legend graphic is available for the layer, this image
           is also retrieved.
        Arguments:
        crs -- coordinate reference: system as a string passed to the WMS
        path_string -- string of waypoints that resemble a vertical
                       section path. Can be omitted for horizontal
                       sections.
        bbox -- bounding box as list of four floats
        width, height -- width and height of requested image in pixels
        """
        # Get layer and style names.
        if not layers:
            layers = [self.multilayers.get_current_layer()]
        if isinstance(layers, Layer):
            layers = [layers]

        layer = layers[0]

        layer_name = [layer.get_layer() for layer in layers]
        style = [layer.get_style() if "image" in format else "" for layer in layers]
        level = layer.get_level_name()

        def normalize_crs(crs):
            if any(crs.startswith(_prefix) for _prefix in ("MSS", "AUTO")):
                return crs.split(",")[0]
            return crs

        if self.check_for_allowed_crs and normalize_crs(crs) not in layer.allowed_crs:
            ret = QtWidgets.QMessageBox.warning(
                self, self.tr("Web Map Service"),
                self.tr(f"WARNING: Selected CRS '{crs}' not contained in allowed list of supported CRS for this WMS\n"
                        f"({layer.allowed_crs})\n"
                        "Continue ?"),
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.Ignore | QtWidgets.QMessageBox.No,
                QtWidgets.QMessageBox.No)
            if ret == QtWidgets.QMessageBox.Ignore:
                self.check_for_allowed_crs = False
            elif ret == QtWidgets.QMessageBox.No:
                return

        # get...Time() will return None if the corresponding checkboxes are
        # disabled. <None> objects passed to wms.getmap will not be included
        # in the query URL that is send to the server.
        init_time = layer.get_itime()
        valid_time = layer.get_vtime()

        logging.debug("fetching layer '%s'; style '%s', width %d, height %d",
                      layer, style, width, height)
        logging.debug("crs=%s, path=%s", crs, path_string)
        logging.debug("init_time=%s, valid_time=%s", init_time, valid_time)
        logging.debug("level=%s", level)
        logging.debug("transparent=%s", transparent)

        try:
            # Call the self.wms.getmap() method in a separate thread to keep
            # the GUI responsive.
            # NOTES:
            # a) This can be done with either a QThread (seems to require
            #    subclassing), or a Python thread (the way I did here).
            #    See: http://stackoverflow.com/questions/1595649/
            #                threading-in-a-pyqt-application-use-qt-threads-or-python-threads
            # b) To retrieve the return value of getmap, a Queue is used.
            #    See: http://stackoverflow.com/questions/1886090/return-value-from-thread
            kwargs = {"layers": layer_name,
                      "styles": style,
                      "srs": crs,
                      "bbox": bbox,
                      "path_str": path_string,
                      "time": valid_time,
                      "init_time": init_time,
                      "level": level,
                      "time_name": layer.vtime_name,
                      "init_time_name": layer.itime_name,
                      "size": (width, height),
                      "format": format,
                      "transparent": transparent}
            legend_kwargs = {"urlstr": layer.get_legend_url(), "md5_filename": None}
            if legend_kwargs["urlstr"] is not None:
                legend_kwargs["md5_filename"] = os.path.join(
                    self.wms_cache, hashlib.md5(legend_kwargs["urlstr"].encode('utf-8')).hexdigest() + ".png")

            # If caching is enabled, get the URL and check the image cache
            # directory for the suitable image file.
            if self.caching_enabled():
                prefetch_config = config_loader(dataset="wms_prefetch")
                prefetch_entries = ["validtime_fwd", "validtime_bck", "level_up", "level_down"]
                for _x in prefetch_entries:
                    if _x in prefetch_config:
                        try:
                            value = int(prefetch_config[_x])
                        except ValueError as ex:
                            logging.error("ERROR: %s %s", type(ex), ex)
                            value = 0
                        prefetch_config[_x] = max(0, value)
                    else:
                        prefetch_config[_x] = 0
                fetch_nr = sum([prefetch_config[_x] for _x in prefetch_entries])
                if fetch_nr > 0:
                    pre_tfwd, pre_tbck, pre_lfwd, pre_lbck = [prefetch_config[_x] for _x in prefetch_entries]
                    ci_time, ci_level = self.cbValidTime.currentIndex(), self.cbLevel.currentIndex()
                    prefetch_key_values = \
                        [("time", self.cbValidTime.itemText(ci_p))
                         for ci_p in list(range(ci_time + 1, ci_time + 1 + pre_tfwd)) +
                            list(range(ci_time - pre_tbck, ci_time))
                         if 0 <= ci_p < self.cbValidTime.count()] + \
                        [("level", self.cbLevel.itemText(ci_p).split(" (")[0])
                         for ci_p in range(ci_level - pre_lbck, ci_level + 1 + pre_lfwd)
                         if ci_p != ci_level and 0 <= ci_p < self.cbLevel.count()]
                    prefetch_maps = []
                    for key, value in prefetch_key_values:
                        kwargs_new = kwargs.copy()
                        kwargs_new[key] = value
                        prefetch_maps.append((layer, kwargs_new, self.get_md5_filename(layer, kwargs_new), True, {}))
                    self.prefetch.emit(prefetch_maps)

            md5_filename = self.get_md5_filename(layer, kwargs)
            self.expected_img = md5_filename
            self.pdlg.reset()
            return [(layer, kwargs, md5_filename, self.caching_enabled(), legend_kwargs)]

        except Exception as ex:
            self.display_exception(ex)

    @QtCore.pyqtSlot(object, object, object, object, object, object, object)
    def continue_retrieve_image(self, img, legend_img, layer, style, init_time, valid_time, md5_filename):
        if self.pdlg.wasCanceled() or self.expected_img != md5_filename:
            return
        self.pdlg.close()
        if img is None:
            return

        complete_level = self.cbLevel.currentText()
        complete_level = complete_level if complete_level != "" else None
        self.display_retrieved_image(img, legend_img, layer, style, init_time, valid_time, complete_level)
        # this is for cases where 'remove' button is clicked, then 'retrieve' is clicked
        self.signal_disable_cbs.emit()
        self.image_displayed.emit()

    def get_map(self, layers=None):
        """Prototypical stub for getMap() function. Needs to be reimplemented
           in derived classes.
        """
        logging.error("getMap not implemented in base class.")

    def after_redraw(self):
        """Event handler that a canvas can call after it has been redrawn.
        """
        self.auto_update()

    def clearCache(self):
        """Clear the image file cache. First ask the user for confirmation.
        """
        # User confirmation to clear the cache.
        clear = (QtWidgets.QMessageBox.question(
            self, "Clear Cache",
            "Do you really want to clear the cache? All stored image "
            "files will be deleted.",
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No) == QtWidgets.QMessageBox.Yes)
        if clear:
            # Delete all files in cache.
            if self.wms_cache is not None:
                cached_files = os.listdir(self.wms_cache)
                logging.debug("clearing cache; deleting %i files...", len(cached_files))
                try:
                    for f in cached_files:
                        os.remove(os.path.join(self.wms_cache, f))
                except (IOError, OSError) as ex:
                    msg = f"ERROR: Cannot delete file '{f}'. ({type(ex)}: {ex})"
                    logging.error(msg)
                    QtWidgets.QMessageBox.critical(self, self.tr("Web Map Service"), self.tr(msg))
                else:
                    logging.debug("cache has been cleared.")
            else:
                logging.debug("no cache exists that can be cleared.")

    def service_cache(self):
        """Service the cache: Remove all files older than the maximum file
           age specified in mss_settings, and remove the oldest files if the
           maximum cache size has been reached.
        """
        logging.debug("servicing cache...")

        # Create a list of all files in the cache.
        files = [os.path.join(self.wms_cache, f) for f in os.listdir(self.wms_cache)]
        # Add the ages of the files (via modification times in sec since epoch)
        # and the file sizes.
        # (current time in sec since epoch)
        current_time = time.time()
        files = [(f, os.path.getsize(f), current_time - os.path.getmtime(f))
                 for f in files if os.path.isfile(f)]
        # Sort the files accordings to their age (i.e. use the
        # third argument of the tuples (file, fsize, fage) as the sorting key).
        files.sort(key=lambda x: x[2])

        # Loop over all cached files, staring with the youngest. Once the
        # maximum cache size has been reached, all files left will be
        # removed (i.e. the oldest files). All files exceeding the maximum
        # file age will also be removed.
        cum_size_bytes = 0
        removed_files = 0
        try:
            for f, fsize, fage in files:
                cum_size_bytes += fsize
                if (cum_size_bytes > config_loader(dataset="wms_cache_max_size_bytes") or
                        fage > config_loader(dataset="wms_cache_max_age_seconds")):
                    os.remove(f)
                    removed_files += 1
        except (IOError, OSError) as ex:
            msg = f"ERROR: Cannot delete file '{f}'. ({type(ex)}: {ex})"
            logging.error(msg)
            QtWidgets.QMessageBox.critical(self, self.tr("Web Map Service"), self.tr(msg))
        logging.debug("cache has been cleaned (%i files removed).", removed_files)

    def squash_multiple_images(self, imgs):
        """
        Lay the images on top of each other from 0 to n and return the product
        """
        background = imgs[0]
        if len(imgs) > 1:
            for foreground in imgs[1:]:
                background.paste(foreground, (0, 0), foreground.convert("RGBA"))

        return background

    def append_multiple_images(self, imgs):
        """
        Stack the list of images vertically and return the product
        """
        images = [x for x in imgs if x]
        if images:
            # Add border around seperate legends
            if len(images) > 1:
                images = [ImageOps.expand(x, border=1, fill="black") for x in images]
            max_height = int((self.view.fig.get_size_inches() * self.view.fig.get_dpi())[1] * 0.99)
            width = max([image.width for image in images])
            height = sum([image.height for image in images])
            result = Image.new("RGBA", (width, height))
            current_height = 0
            for i, image in enumerate(images):
                result.paste(image, (0, current_height - i))
                current_height += image.height

            if max_height < result.height:
                result.thumbnail((result.width, max_height), Image.ANTIALIAS)
            return result

        ################################################################################

#
# CLASS VSecWMSControlWidget
#


class VSecWMSControlWidget(WMSControlWidget):
    """Subclass of WMSControlWidget that extends the WMS client to
       handle (non-standard) vertical sections.
    """

    def __init__(self, parent=None, default_WMS=None, waypoints_model=None,
                 view=None, wms_cache=None):
        super(VSecWMSControlWidget, self).__init__(
            parent=parent, default_WMS=default_WMS, wms_cache=wms_cache, view=view)
        self.waypoints_model = waypoints_model
        self.btGetMap.clicked.connect(self.get_all_maps)

    def get_all_maps(self):
        if self.multilayers.cbMultilayering.isChecked():
            self.get_vsec(self.multilayers.get_active_layers())
        else:
            self.get_vsec([self.multilayers.get_current_layer()])

    def setFlightTrackModel(self, model):
        """Set the QAbstractItemModel instance from which the waypoints
           are obtained.
        """
        self.waypoints_model = model

    @QtCore.Slot()
    def call_get_vsec(self):
        if self.btGetMap.isEnabled() and self.cbAutoUpdate.isChecked() and not self.layerChangeInProgress:
            self.get_all_maps()

    def get_vsec(self, layers=None):
        """Slot that retrieves the vertical section and passes the image
                   to the view.
                """
        # Specify the coordinate reference system for the request. We will
        # use the non-standard "VERT:LOGP" to query the MSS WMS Server for
        # a vertical section. The vertical extent of the plot is specified
        # via the bounding box.
        crs = "VERT:LOGP"
        bbox = self.view.getBBOX()

        # Get lat/lon coordinates of flight track and convert to string for URL.
        path_string = ""
        for waypoint in self.waypoints_model.all_waypoint_data():
            path_string += f"{waypoint.lat:.2f},{waypoint.lon:.2f},"
        path_string = path_string[:-1]

        # Determine the current size of the vertical section plot on the
        # screen in pixels. The image will be retrieved in this size.
        width, height = self.view.get_plot_size_in_px()

        # Retrieve the image.
        if not layers:
            layers = [self.multilayers.get_current_layer()]
        layers.sort(key=lambda x: self.multilayers.get_multilayer_priority(x))

        args = []
        for i, layer in enumerate(layers):
            transparent = self.cbTransparent.isChecked() if i == 0 else True
            args.extend(self.retrieve_image(layer, crs, bbox, path_string, width, height, transparent))

        self.fetch.emit(args)

    def display_retrieved_image(self, imgs, legend_imgs, layer, style, init_time, valid_time, level):
        # Plot the image on the view canvas.
        self.view.draw_image(self.squash_multiple_images(imgs))
        self.view.draw_legend(self.append_multiple_images(legend_imgs))
        style_title = self.multilayers.get_current_layer().get_style()
        self.view.draw_metadata(title=self.multilayers.get_current_layer().layerobj.title,
                                init_time=init_time,
                                valid_time=valid_time,
                                style=style_title)

    def is_layer_aligned(self, layer):
        crss = getattr(layer, "crsOptions", None)
        return crss is not None and any(crs.startswith("VERT") for crs in crss)

#
# CLASS HSecWMSControlWidget
#


class HSecWMSControlWidget(WMSControlWidget):
    """Subclass of WMSControlWidget that extends the WMS client to
       handle (standard) horizontal sections, i.e. maps.
    """

    def __init__(self, parent=None, default_WMS=None, view=None, wms_cache=None):
        super(HSecWMSControlWidget, self).__init__(
            parent=parent, default_WMS=default_WMS, wms_cache=wms_cache, view=view)
        self.btGetMap.clicked.connect(self.get_all_maps)

    def level_changed(self):
        super().level_changed()
        self.view.waypoints_interactor.update()

    def get_map(self, layers=None):
        """Slot that retrieves the map and passes the image
           to the view.
        """
        # Get coordinate reference system and bounding box from the map
        # object in the view.
        crs = self.view.get_crs()
        bbox = self.view.getBBOX()

        # Determine the current size of the vertical section plot on the
        # screen in pixels. The image will be retrieved in this size.
        width, height = self.view.get_plot_size_in_px()

        if not layers:
            layers = [self.multilayers.get_current_layer()]

        layers.sort(key=lambda x: self.multilayers.get_multilayer_priority(x))

        args = []
        for i, layer in enumerate(layers):
            transparent = self.cbTransparent.isChecked() if i == 0 else True
            bbox_tmp = tuple(bbox)
            wms = self.multilayers.layers[layer.wms_name]["wms"]
            if wms.version == "1.3.0" and crs.startswith("EPSG") and int(crs[5:]) in axisorder_yx:
                bbox_tmp = (bbox[1], bbox[0], bbox[3], bbox[2])
            args.extend(self.retrieve_image(layer, crs, bbox_tmp, None, width, height, transparent))

        self.fetch.emit(args)

    def display_retrieved_image(self, imgs, legend_imgs, layer, style, init_time, valid_time, level):
        # Plot the image on the view canvas.
        style_title = self.multilayers.get_current_layer().get_style()
        self.view.draw_metadata(title=self.multilayers.get_current_layer().layerobj.title,
                                init_time=init_time,
                                valid_time=valid_time,
                                level=level,
                                style=style_title)
        self.view.draw_image(self.squash_multiple_images(imgs))
        self.view.draw_legend(self.append_multiple_images(legend_imgs))
        self.view.waypoints_interactor.update()

    def is_layer_aligned(self, layer):
        crss = getattr(layer, "crsOptions", None)
        return crss is not None and not any(crs.startswith("VERT") for crs in crss) \
            and not any(crs.startswith("LINE") for crs in crss)


class LSecWMSControlWidget(WMSControlWidget):
    """Subclass of WMSControlWidget that extends the WMS client to
       handle (non-standard) linear sections.
    """

    def __init__(self, parent=None, default_WMS=None, waypoints_model=None,
                 view=None, wms_cache=None):
        super(LSecWMSControlWidget, self).__init__(
            parent=parent, default_WMS=default_WMS, wms_cache=wms_cache, view=view)
        self.waypoints_model = waypoints_model
        self.btGetMap.clicked.connect(self.get_all_maps)

    def get_all_maps(self):
        if self.multilayers.cbMultilayering.isChecked():
            self.get_lsec(self.multilayers.get_active_layers())
        else:
            self.get_lsec([self.multilayers.get_current_layer()])

    def setFlightTrackModel(self, model):
        """Set the QAbstractItemModel instance from which the waypoints
           are obtained.
        """
        self.waypoints_model = model

    @QtCore.Slot()
    def call_get_lsec(self):
        if self.btGetMap.isEnabled() and self.cbAutoUpdate.isChecked() and not self.layerChangeInProgress:
            self.get_all_maps()

    def get_lsec(self, layers=None):
        """Slot that retrieves the linear plot and passes the image
                   to the view.
                """
        # Specify the coordinate reference system for the request. We will
        # use the non-standard "LINE:1" to query the MSS WMS Server for
        # a linear plot.
        crs = "LINE:1"
        bbox = self.view.getBBOX()

        # Get lat/lon/alt coordinates of flight track and convert to string for URL.
        path_string = ""
        for waypoint in self.waypoints_model.all_waypoint_data():
            path_string += f"{waypoint.lat:.2f},{waypoint.lon:.2f},{waypoint.pressure},"
        path_string = path_string[:-1]

        # Retrieve the image.
        if not layers:
            layers = [self.multilayers.get_current_layer()]
        layers.sort(key=lambda x: self.multilayers.get_multilayer_priority(x))

        args = []
        for i, layer in enumerate(layers):
            args.extend(self.retrieve_image(layer, crs, bbox, path_string, transparent=False, format="text/xml"))

        self.fetch.emit(args)

    def display_retrieved_image(self, imgs, legend_imgs, layer, style, init_time, valid_time, level):
        # Plot the image on the view canvas.
        layers = self.multilayers.get_active_layers() if self.multilayers.cbMultilayering.isChecked() else \
            [self.multilayers.get_current_layer()]
        layers.sort(key=lambda x: self.multilayers.get_multilayer_priority(x))
        colors = [layer.color for layer in layers]
        scales = [layer.get_style() for layer in layers]
        self.view.draw_image(imgs, colors, scales)
        self.view.draw_legend(self.append_multiple_images(legend_imgs))
        style_title = self.multilayers.get_current_layer().get_style()
        self.view.draw_metadata(title=self.multilayers.get_current_layer().layerobj.title,
                                init_time=init_time,
                                valid_time=valid_time,
                                style=style_title)

    def is_layer_aligned(self, layer):
        crss = getattr(layer, "crsOptions", None)
        return crss is not None and any(crs.startswith("LINE") for crs in crss)
